"""
Halo filters to be used with the HaloProfiler.

Author: Britton Smith <brittons@origins.colorado.edu>
Affiliation: CASA/University of CO, Boulder
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2011 Britton Smith.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from copy import deepcopy
import numpy as np

from yt.funcs import *

def VirialFilter(profile, overdensity_field='ActualOverdensity',
                 virial_overdensity=200., must_be_virialized=True,
                 virial_filters=[['TotalMassMsun', '>=','1e14']],
                 virial_quantities=['TotalMassMsun', 'RadiusMpc'],
                 virial_index=None, use_log=False):
    r"""Filter halos by virial quantities.
    
    Return values are a True or False whether the halo passed the filter, 
    along with a dictionary of virial quantities for the fields specified in 
    the virial_quantities keyword.  Thresholds for virial quantities are 
    given with the virial_filters keyword in the following way:
    [field, condition, value].
    
    This is typically used as part of a call to `add_halo_filter`.
    
    Parameters
    ----------
    overdensity_field : string
        The field used for interpolation with the 
        specified critical value given with 'virial_overdensity'.  
        Default='ActualOverdensity'.
    virial_overdensity : float
        The value used to determine the outer radius of the virialized halo.
        Default: 200.
    must_be_virialized : bool
        If no values in the profile are above the 
        value of virial_overdensity, the halo does not pass the filter.  
        Default: True.
    virial_filters : array_like
        Conditional filters based on virial quantities 
        given in the following way: [field, condition, value].  
        Default: [['TotalMassMsun', '>=','1e14']].
    virial_quantities : array_like
        Fields for which interpolated values should 
        be calculated and returned.  Default: ['TotalMassMsun', 'RadiusMpc'].
    virial_index : array_like
        If given as a list, the index of the radial profile 
        which is used for interpolation is placed here.  Default: None.
    use_log : bool
        If True, interpolation is done in log space.  
        Default: False.
    
    Examples
    --------
    >>> hp.add_halo_filter(HP.VirialFilter, must_be_virialized=True,
                   overdensity_field='ActualOverdensity',
                   virial_overdensity=200,
                   virial_filters=[['TotalMassMsun','>=','1e14']],
                   virial_quantities=['TotalMassMsun','RadiusMpc'])
    
    """

    fields = deepcopy(virial_quantities)
    if virial_filters is None: virial_filters = []
    for vfilter in virial_filters:
        if not vfilter[0] in fields:
            fields.append(vfilter[0])
    
    overDensity = []
    temp_profile = dict((field, []) for field in fields)

    for q in range(len(profile[overdensity_field])):
        good = True
        if (profile[overdensity_field][q] != profile[overdensity_field][q]):
            good = False
            continue
        for field in fields:
            if (profile[field][q] != profile[field][q]):
                good = False
                break
        if good:
            overDensity.append(profile[overdensity_field][q])
            for field in fields:
                temp_profile[field].append(profile[field][q])

    if use_log:
        for field in temp_profile.keys():
            temp_profile[field] = np.log10(np.clip(temp_profile[field], 1e-90, 
                                                   max(temp_profile[field])))

    virial = dict((field, 0.0) for field in fields)

    if (not (np.array(overDensity) >= virial_overdensity).any()) and \
            must_be_virialized:
        mylog.debug("This halo is not virialized!")
        return [False, {}]

    if (len(overDensity) < 2):
        mylog.debug("Skipping halo with no valid points in profile.")
        return [False, {}]

    if (overDensity[1] <= virial_overdensity):
        index = 0
    elif (overDensity[-1] >= virial_overdensity):
        index = -2
    else:
        for q in (np.arange(len(overDensity),0,-1)-1):
            if (overDensity[q] < virial_overdensity) and (overDensity[q-1] >= virial_overdensity):
                index = q - 1
                break

    if type(virial_index) is list:
        virial_index.append(index)

    for field in fields:
        if (overDensity[index+1] - overDensity[index]) == 0:
            mylog.debug("Overdensity profile has slope of zero.")
            return [False, {}]
        else:
            slope = (temp_profile[field][index+1] - temp_profile[field][index]) / \
                (overDensity[index+1] - overDensity[index])
            value = slope * (virial_overdensity - overDensity[index]) + \
                temp_profile[field][index]
            virial[field] = value

    if use_log:
        for field in virial.keys():
            virial[field] = np.power(10, virial[field])

    for vfilter in virial_filters:
        if eval("%s %s %s" % (virial[vfilter[0]],vfilter[1],vfilter[2])):
            mylog.debug("(%s %s %s) returned True for %s." % \
                            (vfilter[0],vfilter[1],vfilter[2],virial[vfilter[0]]))
            continue
        else:
            mylog.debug("(%s %s %s) returned False for %s." % \
                            (vfilter[0],vfilter[1],vfilter[2],virial[vfilter[0]]))
            return [False, {}]

    return [True, dict((("%s_%s" % (q, virial_overdensity)), virial[q])
                       for q in virial_quantities)]




"""
Testsuite for PlotWindow class

Author: Nathan Goldbaum <goldbaum@ucolick.org>
Affiliation: UCSC Astronomy
Homepage: http://yt-project.org/
License:
  Copyright (C) 2012 Nathan Goldbaum.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import tempfile
import shutil
from yt.testing import \
    fake_random_pf, assert_equal, assert_rel_equal
from yt.utilities.answer_testing.framework import \
    requires_pf, data_dir_load, PlotWindowAttributeTest
from yt.visualization.api import \
    SlicePlot, ProjectionPlot, OffAxisSlicePlot, OffAxisProjectionPlot


def setup():
    """Test specific setup."""
    from yt.config import ytcfg
    ytcfg["yt", "__withintesting"] = "True"


def assert_fname(fname):
    """Function that checks file type using libmagic"""
    if fname is None:
        return

    with open(fname, 'rb') as fimg:
        data = fimg.read()
    data = str(data)
    image_type = ''

    # see http://www.w3.org/TR/PNG/#5PNG-file-signature
    if data.startswith('\211PNG\r\n\032\n'):
        image_type = '.png'
    # see http://www.mathguide.de/info/tools/media-types/image/jpeg
    elif data.startswith('\377\330'):
        image_type = '.jpeg'
    elif data.startswith('%!PS-Adobe'):
        if 'EPSF' in data[:data.index('\n')]:
            image_type = '.eps'
        else:
            image_type = '.ps'
    elif data.startswith('%PDF'):
        image_type = '.pdf'

    return image_type == os.path.splitext(fname)[1]

attr_args ={ "pan"             : [( ((0.1, 0.1),), {} )],
             "pan_rel"         : [( ((0.1, 0.1),), {} )],
             "set_axes_unit"   : [( ("kpc",), {} ),
                                  ( ("Mpc",), {} ),
                                  ( (("kpc", "kpc"),), {} ),
                                  ( (("kpc", "Mpc"),), {} )],
             "set_buff_size"   : [( (1600,), {} ),
                                  ( ((600, 800),), {} )],
             "set_center"      : [( ((0.4, 0.3),), {} )],
             "set_cmap"        : [( ('Density', 'RdBu'), {} ),
                                  ( ('Density', 'kamae'), {} )],
             "set_font"        : [( ({'family':'sans-serif', 'style':'italic',
                                      'weight':'bold', 'size':24},), {} )],
             "set_log"         : [( ('Density', False), {} )],
             "set_window_size" : [( (7.0,), {} )],
             "set_zlim" : [( ('Density', 1e-25, 1e-23), {} ),
                           ( ('Density', 1e-25, None), {'dynamic_range' : 4} )],
             "zoom" : [( (10,), {} )] }

m7 = "DD0010/moving7_0010"
wt = "WindTunnel/windtunnel_4lev_hdf5_plt_cnt_0030"
@requires_pf(m7)
@requires_pf(wt)
def test_attributes():
    """Test plot member functions that aren't callbacks"""
    plot_field = 'Density'
    decimals = 3

    pf = data_dir_load(m7)
    for ax in 'xyz':
        for attr_name in attr_args.keys():
            for args in attr_args[attr_name]:
                yield PlotWindowAttributeTest(pf, plot_field, ax, attr_name,
                                              args, decimals)
    pf = data_dir_load(wt)
    ax = 'z'
    for attr_name in attr_args.keys():
        for args in attr_args[attr_name]:
            yield PlotWindowAttributeTest(pf, plot_field, ax, attr_name,
                                          args, decimals)

def test_setwidth():
    pf = fake_random_pf(64)

    slc = SlicePlot(pf, 0, 'Density')

    yield assert_equal, [slc.xlim, slc.ylim, slc.width], \
        [(0.0, 1.0), (0.0, 1.0), (1.0, 1.0)]

    slc.set_width((0.5,0.8))

    yield assert_rel_equal, [slc.xlim, slc.ylim, slc.width], \
        [(0.25, 0.75), (0.1, 0.9), (0.5, 0.8)], 15

    slc.set_width(15,'kpc')

    yield assert_rel_equal, [slc.xlim, slc.ylim, slc.width], \
        [(-7.5/pf['kpc'], 7.5/pf['kpc']),
         (-7.5/pf['kpc'], 7.5/pf['kpc']),
         (15/pf['kpc'], 15/pf['kpc'])], 15

    slc.set_width((15,'kpc'))

    yield assert_rel_equal, [slc.xlim, slc.ylim, slc.width], \
        [(-7.5/pf['kpc'], 7.5/pf['kpc']),
         (-7.5/pf['kpc'], 7.5/pf['kpc']),
         (15/pf['kpc'], 15/pf['kpc'])], 15

    slc.set_width(((15,'kpc'),(10,'kpc')))

    yield assert_rel_equal, [slc.xlim, slc.ylim, slc.width], \
        [(-7.5/pf['kpc'], 7.5/pf['kpc']),
         (-5/pf['kpc'], 5/pf['kpc']),
         (15/pf['kpc'], 10/pf['kpc'])], 15

    slc.set_width(((15,'kpc'),(10000,'pc')))

    yield assert_rel_equal, [slc.xlim, slc.ylim, slc.width], \
        [(-7.5/pf['kpc'], 7.5/pf['kpc']),
         (-5/pf['kpc'], 5/pf['kpc']),
         (15/pf['kpc'], 10/pf['kpc'])], 15

    slc.set_width((15,'kpc'),(10000,'pc'))

    yield assert_rel_equal, [slc.xlim, slc.ylim, slc.width], \
        [(-7.5/pf['kpc'], 7.5/pf['kpc']),
         (-5/pf['kpc'], 5/pf['kpc']),
         (15/pf['kpc'], 10/pf['kpc'])], 15

def test_save():
    """Test plot window creation and saving to disk."""
    # Perform I/O in safe place instead of yt main dir
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    normal = [1, 1, 1]

    test_pf = fake_random_pf(64)
    test_flnms = [None, 'test.png', 'test.eps',
                  'test.ps', 'test.pdf']

    ds_region = test_pf.h.region([0.5]*3,[0.4]*3,[0.6]*3)

    for dim in [0, 1, 2]:
        obj = SlicePlot(test_pf, dim, 'Density')
        for fname in test_flnms:
            yield assert_equal, assert_fname(obj.save(fname)[0]), True

    for dim in [0, 1, 2]:
        obj = ProjectionPlot(test_pf, dim, 'Density')
        for fname in test_flnms:
            yield assert_equal, assert_fname(obj.save(fname)[0]), True
        # Test ProjectionPlot's data_source keyword
        obj = ProjectionPlot(test_pf, dim, 'Density',
                             data_source=ds_region)
        obj.save()

    obj = OffAxisSlicePlot(test_pf, normal, 'Density')
    for fname in test_flnms:
        yield assert_equal, assert_fname(obj.save(fname)[0]), True

    obj = OffAxisProjectionPlot(test_pf, normal, 'Density')
    for fname in test_flnms:
        yield assert_equal, assert_fname(obj.save(fname)[0]), True

    os.chdir(curdir)
    # clean up
    shutil.rmtree(tmpdir)



#!/usr/bin/env python
from os.path import join


def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration, get_numpy_include_dirs
    from numpy.distutils.system_info import get_info
    from distutils.sysconfig import get_python_inc

    config = Configuration('spatial', parent_package, top_path)

    config.add_data_dir('tests')

#    qhull_src = ['geom2.c', 'geom.c', 'global.c', 'io.c', 'libqhull.c',
#                 'mem.c', 'merge.c', 'poly2.c', 'poly.c', 'qset.c',
#                 'random.c', 'rboxlib.c', 'stat.c', 'user.c', 'usermem.c',
#                 'userprintf.c']

#    config.add_library('qhull',
#                       sources=[join('qhull', 'src', x) for x in qhull_src],
#                       include_dirs=[get_python_inc(),
#                                     get_numpy_include_dirs()],
#                       # XXX: GCC dependency!
#                       #extra_compiler_args=['-fno-strict-aliasing'],
#                       )

#    lapack = dict(get_info('lapack_opt'))
#    try:
#        libs = ['qhull'] + lapack.pop('libraries')
#    except KeyError:
#        libs = ['qhull']
#    config.add_extension('qhull',
#                         sources=['qhull.c'],
#                         libraries=libs,
#                         **lapack)

    config.add_extension('ckdtree', sources=['ckdtree.pyx'],
        libraries=["m"],
        include_dirs=[get_numpy_include_dirs()])

    config.add_extension('_distance_wrap',
        sources=[join('src', 'distance_wrap.c'), join('src', 'distance.c')],
        include_dirs=[get_numpy_include_dirs()])

    return config

if __name__ == '__main__':
    from numpy.distutils.core import setup
    setup(maintainer="SciPy Developers",
          author="Anne Archibald",
          maintainer_email="scipy-dev@scipy.org",
          description="Spatial algorithms and data structures",
          url="http://www.scipy.org",
          license="SciPy License (BSD Style)",
          **configuration(top_path='').todict()
          )



"""
Test for off_axis_projection and write_projection

Author: Cameron Hummels <chummels@gmail.com>
Affiliation: University of Arizona
Homepage: http://yt-project.org/
License:
  Copyright (C) 2013 Cameron Hummels.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import os.path
import tempfile
import shutil
from yt.testing import \
    fake_random_pf, assert_equal, expand_keywords
from yt.mods import \
    off_axis_projection, write_projection


def setup():
    """Test specific setup."""
    from yt.config import ytcfg
    ytcfg["yt", "__withintesting"] = "True"


def test_write_projection():
    """Tests functionality of off_axis_projection and write_projection."""
    # Perform I/O in safe place instead of yt main dir
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    # args for off_axis_projection
    test_pf = fake_random_pf(64)
    c = [0.5, 0.5, 0.5]
    norm = [0.5, 0.5, 0.5]
    W = [0.5,0.5,1.0]
    N = 64
    field = "Density"
    oap_args = [test_pf, c, norm, W, N, field]

    # kwargs for off_axis_projection
    oap_kwargs = {}
    oap_kwargs['weight'] = (None, 'CellMassMsun')
    oap_kwargs['no_ghost'] = (True, False)
    oap_kwargs['interpolated'] = (True, False)
    oap_kwargs['north_vector'] = ((1,0,0), (0,0.5,1.0))
    oap_kwargs_list = expand_keywords(oap_kwargs)

    # args for write_projection
    fn = "test.png"

    # kwargs for write_projection
    wp_kwargs = {}
    wp_kwargs['colorbar'] = (True, False)
    wp_kwargs['colorbar_label'] = ('test')
    wp_kwargs['title'] = ('test')
    wp_kwargs['limits'] = (None, (1e3, 1e5))
    wp_kwargs['take_log'] = (True, False)
    wp_kwargs['figsize'] = ((8,6), [1,1])
    wp_kwargs['dpi'] = (100, 50)
    wp_kwargs['cmap_name'] = ('algae', 'jet')
    wp_kwargs_list = expand_keywords(wp_kwargs)

    # test all off_axis_projection kwargs and write_projection kwargs
    # make sure they are able to be projected, then remove and try next
    # iteration
    for oap_kwargs in oap_kwargs_list:
        image = off_axis_projection(*oap_args, **oap_kwargs)
        for wp_kwargs in wp_kwargs_list:
            write_projection(image, fn, **wp_kwargs)
            yield assert_equal, os.path.exists(fn), True
            os.remove(fn)

    os.chdir(curdir)
    # clean up
    shutil.rmtree(tmpdir)



"""
Compatibility module

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from .CICDeposit import *
from .ContourFinding import *
from .DepthFirstOctree import *
from .fortran_reader import *
from .freetype_writer import *
from .geometry_utils import *
from .Interpolators import *
from .misc_utilities import *
from .Octree import *
from .png_writer import *
from .PointsInVolume import *
from .QuadTree import *
from .RayIntegrators import *
from .grid_traversal import *
from .marching_cubes import *
from .GridTree import *
from .write_array import *



"""
A first-pass at a VisVis-based interactive viewer.

Hi Matt,

You could define your MultiTexture object (which would be a Wobject) and use
multiple visvis.textures.TextureObject instances to represent the sub-textures.
This TextureObject class is not a Wobject, but simply of a wrapper around an
OpenGl texture that handles uploading, updating, etc. 

The MultiTexture object you could derive from BaseTexture or maybe even
Texture3D if your data is 3D. Or maybe just from Wobject if this makes more
sense. Also, the visvis.textures.TextureObjectToVisualize might be used instead
of the TextureObject class.

I mentioned a lot of texture classes, let's make a small list:
  * TextureObject - wraps an OpenGl texture
  * TextureObjectToVisualize(TextureObject) - implements intensity scaling
    (what is changed with the clim property)
  * BaseTexture(Wobject) - the base visvis Wobject class to represent textures.
    It *contains* one TextureObjectToVisualize and has several properties to
    influence it's appearance.  * Texture2D(BaseTexture) - Implementation for 2D
    textures.
  * Texture3D(BaseTexture) - Implementation for 3D textures.
  * YourNewMultiTexture(BaseTexture or ...) - contains multiple
    TextureObjectToVisualize or TextureObject instances.

This is I think the most sensible way to go about this. 

Now in the OnDraw method, you can draw the textures one by one using the same
settings. Or if you want a single draw in which you combine the information
from all textures to determine the final color, you should write a new glsl
shader to which you pass all textures.

I hope this helps,
  Almar
"""
from yt.mods import *
from yt.funcs import *

import visvis as vv
import visvis.textures as vvt

import OpenGL.GL as gl
import OpenGL.GL.ARB.shader_objects as gla
import OpenGL.GLU as glu

import numpy as np

class MultipleTexture(vv.Wobject):
    def __init__(self, parent, data, global_size, renderStyle='mip'):
        vv.Wobject.__init__(self, parent)

        self._global_size = global_size
        
        # create colormap
        self._colormap = vvt.Colormap()
        
        # create glsl program for this texture...
        self._program1 = program =  vvt.GlslProgram()
        
        # scale and translation transforms
        self._trafo_scale = vv.Transform_Scale()
        self._trafo_trans = vv.Transform_Translate()
        self.transformations.append(self._trafo_trans)
        self.transformations.append(self._trafo_scale)        

        data = ensure_list(data)
        self._textures = []
        self._quads = {}
        for obj in data:
            tex = vvt.TextureObjectToVisualize(3, obj) 
            self._textures.append(tex)
            tex.SetData(obj)
        self._program1.SetVertexShader(vvt.vshaders['calculateray'])
        self._program1.SetFragmentShader(vvt.fshaders['mip'])
        self._renderStyle = ''
        self.renderStyle = renderStyle
        if not self._renderStyle:
            self.renderStyle = 'mip'
        self._quadlists = None

    def OnDrawShape(self, clr):
        gl.glColor(clr[0], clr[1], clr[2], 1.0)
        self._DrawQuads()

    def OnDraw(self, fast=False):
        # Prepare by setting things to their defaults. This might release some
        # memory so result in a bigger chance that the shader is run in 
        # hardware mode. On ATI, the line and point smoothing should be off
        # if you want to use gl_FragCoord. (Yeah, I do not see the connection
        # either...)
        gl.glPointSize(1)
        gl.glLineWidth(1)
        gl.glDisable(gl.GL_LINE_STIPPLE)
        gl.glDisable(gl.GL_LINE_SMOOTH)
        gl.glDisable(gl.GL_POINT_SMOOTH)
        
        # only draw front-facing parts
        gl.glEnable(gl.GL_CULL_FACE)
        gl.glCullFace(gl.GL_BACK)
        gl.glBlendFunc(gl.GL_ONE, gl.GL_ONE)
        gl.glBlendEquation(gl.GL_MAX)
        gl.glDisable(gl.GL_DEPTH_TEST)
        
        if self._program1.IsUsable():
            self._program1.Enable()

        if fast:
            self._program1.SetUniformf('stepRatio', [0.4])
        else:
            self._program1.SetUniformf('stepRatio', [1.0])

        self._program1.SetUniformi('texture', [0])        
        self._colormap.Enable(1)
        self._program1.SetUniformi('colormap', [1])

        # enable this texture
        t1 = time.time()
        for i,tex in enumerate(self._textures):

            tex.Enable(0)
            
            if not tex._shape:
                continue

            # fragment shader on
                
            # bind texture- and help-textures (create if it does not exist)
            
            # set uniforms: parameters
            shape = tex._shape # as in opengl
            self._program1.SetUniformf('shape',reversed(list(shape)) )
            
            # do the actual drawing

            self._DrawQuads(tex, i)
            tex.Disable()
        
        # clean up
        gl.glFlush()
        t2 = time.time()
        print "Rendering: %0.3e" % (t2-t1)
        self._colormap.Disable()
        self._program1.Disable()
        #
        gl.glBlendEquation(gl.GL_FUNC_ADD)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)
        gl.glDisable(gl.GL_CULL_FACE)
        gl.glEnable(gl.GL_LINE_SMOOTH)
        gl.glEnable(gl.GL_POINT_SMOOTH)
        gl.glEnable(gl.GL_DEPTH_TEST)
        
        
    def _DrawQuads(self, tex, tex_id):
        """ Draw the quads of the texture. 
        This is done in a seperate method to reuse code in 
        OnDraw() and OnDrawShape(). """        
        
        # should we draw?
        if not tex._shape:
            return 
        
        # should we create quads?
        if tex_id not in self._quads:
            self._CreateQuads(tex, tex_id)
        
        # get data
        tex_coord, ver_coord, ind = self._quads[tex_id]
        
        # init vertex and texture array
        gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
        gl.glEnableClientState(gl.GL_TEXTURE_COORD_ARRAY)
        gl.glVertexPointerf(ver_coord.data)
        gl.glTexCoordPointerf(tex_coord.data)
        
        # draw
        gl.glDrawElements(gl.GL_QUADS, len(ind), gl.GL_UNSIGNED_BYTE, ind)
        
        # disable vertex array        
        gl.glDisableClientState(gl.GL_VERTEX_ARRAY)
        gl.glDisableClientState(gl.GL_TEXTURE_COORD_ARRAY)
    
    
    def _GetLimits(self):
        """ _GetLimits()
        Get the limits in world coordinates between which the object exists.
        """
        
        x1, x2 = -0.5, -0.5 + self._global_size[2]
        y1, y2 = -0.5, -0.5 + self._global_size[1]
        z1, z2 = -0.5, -0.5 + self._global_size[0]

        return vv.Wobject._GetLimits(self, x1, x2, y1, y2, z1, z2)
    
    
    def _CreateQuads(self, tex, tex_id):
        axes = self.GetAxes()
        if not axes:
            return
        
        # Store daspect so we can detect it changing
        self._daspectStored = axes.daspect
        
        # Note that we could determine the world coordinates and use
        # them directly here. However, the way that we do it now (using
        # the transformations) is to be preferred, because that way the
        # transformations are applied via the ModelView matrix stack,
        # and can easily be made undone in the raycaster.
        # The -0.5 offset is to center pixels/voxels. This works correctly
        # for anisotropic data.

        dr = tex._dataRef

        ss = [s - 1 for s in tex._shape]

        x0 = dr.origin[2] * self._global_size[2] - 0.5
        x1 = x0 + ss[2] * dr.sampling[2] * self._global_size[2]

        y0 = dr.origin[1] * self._global_size[1]
        y1 = y0 + ss[1] * dr.sampling[1] * self._global_size[1]

        z0 = dr.origin[0] * self._global_size[0]
        z1 = z0 + ss[0] * dr.sampling[0] * self._global_size[0]
        
        # prepare texture coordinates
        t0, t1 = 0, 1
        
        # if any axis are flipped, make sure the correct polygons are front
        # facing
        tmp = 1
        for i in axes.daspect:
            if i<0:
                tmp*=-1        
        if tmp==1:
            t0, t1 = t1, t0
            x0, x1 = x1, x0
            y0, y1 = y1, y0
            z0, z1 = z1, z0

        # using glTexCoord* is the same as glMultiTexCoord*(GL_TEXTURE0)
        # Therefore we need to bind the base texture to 0.
        
        # draw. So we draw the six planes of the cube (well not a cube,
        # a 3d rectangle thingy). The inside is only rendered if the 
        # vertex is facing front, so only 3 planes are rendered at a        
        # time...                
        
        tex_coord, ver_coord = vvt.points.Pointset(3), vvt.points.Pointset(3)
        indices = [0,1,2,3, 4,5,6,7, 3,2,6,5, 0,4,7,1, 0,3,5,4, 1,7,6,2]
        
        # bottom
        tex_coord.Append((t0,t0,t0)); ver_coord.Append((x0, y0, z0)) # 0
        tex_coord.Append((t1,t0,t0)); ver_coord.Append((x1, y0, z0)) # 1
        tex_coord.Append((t1,t1,t0)); ver_coord.Append((x1, y1, z0)) # 2
        tex_coord.Append((t0,t1,t0)); ver_coord.Append((x0, y1, z0)) # 3
        # top
        tex_coord.Append((t0,t0,t1)); ver_coord.Append((x0, y0, z1)) # 4    
        tex_coord.Append((t0,t1,t1)); ver_coord.Append((x0, y1, z1)) # 5
        tex_coord.Append((t1,t1,t1)); ver_coord.Append((x1, y1, z1)) # 6
        tex_coord.Append((t1,t0,t1)); ver_coord.Append((x1, y0, z1)) # 7
        
        # Store quads
        self._quads[tex_id] = (tex_coord, ver_coord, np.array(indices,dtype=np.uint8))

def visvis_plot(vp):
    """
    This function accepts a volume rendering object, which it then tosses into
    visvis for plotting.
    """
        
    vp.partition_grids()
    gs = vp.bricks

    mi = min((g.my_data[0].min() for g in gs))
    ma = max((g.my_data[0].max() for g in gs))

    texes = []

    ax = vv.gca()

    for i,g in enumerate(gs):
        ss = ((g.RightEdge - g.LeftEdge) / (np.array(g.my_data[0].shape)-1)).tolist()
        origin = g.LeftEdge.astype("float32").tolist()
        dd = (g.my_data[0].astype("float32") - mi)/(ma - mi)
        dd = np.clip(dd, 0.0, 1.0)
        print ss
        texes.append(vv.Aarray(dd, origin = origin, sampling = ss))

    mtex = MultipleTexture(ax, texes, global_size=vp.pf.domain_dimensions)

    ax.daspectAuto = False
    ax.SetLimits()
    ax.bgcolor = (0,0,0)

    # set camera
    ax.cameraType = '3d'

    # done
    ax.Draw()

    return mtex, ax



"""
This module contains a near-replacement for enzo_anyl

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os

from yt.data_objects.profiles import BinnedProfile1D
from yt.funcs import *

class StandardRadialAnalysis(object):
    def __init__(self, pf, center, radius, n_bins = 128, inner_radius = None):
        self.pf = pf
        # We actually don't want to replicate the handling of setting the
        # center here, so we will pass it to the sphere creator.
        # Note also that the sphere can handle (val, unit) for radius, so we
        # will grab that from the sphere as well
        self.obj = pf.h.sphere(center, radius)
        if inner_radius is None:
            inner_radius = pf.h.get_smallest_dx() * pf['cm']
        self.inner_radius = inner_radius
        self.outer_radius = self.obj.radius * pf['cm']
        self.n_bins = n_bins

    def setup_field_parameters(self):
        # First the bulk velocity
        bv = self.obj.quantities["BulkVelocity"]()
        self.obj.set_field_parameter("bulk_velocity", bv)

    def construct_profile(self):
        # inner_bound in cm, outer_bound in same
        # Note that in some cases, we will need to massage this object.
        prof = BinnedProfile1D(self.obj, self.n_bins, "Radius",
                               self.inner_radius, self.outer_radius,
                               lazy_reader = True)
        by_weights = defaultdict(list)
        for fspec in analysis_field_list:
            if isinstance(fspec, types.TupleType) and len(fspec) == 2:
                field, weight = fspec
            else:
                field, weight = fspec, "CellMassMsun"
            by_weights[weight].append(field)
        known_fields = set(self.pf.h.field_list + self.pf.h.derived_field_list)
        for weight, fields in by_weights.items():
            fields = set(fields)
            fields.intersection_update(known_fields)
            prof.add_fields(list(fields), weight=weight)
        self.prof = prof

    def plot_everything(self, dirname = None):
        if not dirname:
            dirname = "%s_profile_plots/" % (self.pf)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)
        import matplotlib; matplotlib.use("Agg")
        import pylab
        for field in self.prof.keys():
            if field in ("UsedBins", "Radius"): continue
            pylab.clf()
            pylab.loglog(self.prof["Radius"], self.prof[field], '-x')
            pylab.xlabel("Radius [cm]")
            pylab.ylabel("%s" % field)
            pylab.savefig(os.path.join(
                dirname, "Radius_%s.png" % (field.replace(" ","_"))))



"""
A means of running standalone commands with a shared set of options.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.config import ytcfg
ytcfg["yt","__command_line"] = "True"
from yt.startup_tasks import parser, subparsers
from yt.mods import *
from yt.funcs import *
from yt.utilities.minimal_representation import MinimalProjectDescription
import argparse, os, os.path, math, sys, time, subprocess, getpass, tempfile
import urllib, urllib2, base64, os

def _fix_pf(arg):
    if os.path.isdir("%s" % arg) and \
        os.path.exists("%s/%s" % (arg,arg)):
        pf = load("%s/%s" % (arg,arg))
    elif os.path.isdir("%s.dir" % arg) and \
        os.path.exists("%s.dir/%s" % (arg,arg)):
        pf = load("%s.dir/%s" % (arg,arg))
    elif arg.endswith(".hierarchy"):
        pf = load(arg[:-10])
    else:
        pf = load(arg)
    return pf

def _add_arg(sc, arg):
    if isinstance(arg, types.StringTypes):
        arg = _common_options[arg].copy()
    argc = dict(arg.items())
    argnames = []
    if "short" in argc: argnames.append(argc.pop('short'))
    if "long" in argc: argnames.append(argc.pop('long'))
    sc.add_argument(*argnames, **argc)

class YTCommand(object):
    args = ()
    name = None
    description = ""
    aliases = ()
    npfs = 1

    class __metaclass__(type):
        def __init__(cls, name, b, d):
            type.__init__(cls, name, b, d)
            if cls.name is not None:
                names = ensure_list(cls.name)
                for name in names:
                    sc = subparsers.add_parser(name,
                        description = cls.description,
                        help = cls.description)
                    sc.set_defaults(func=cls.run)
                    for arg in cls.args:
                        _add_arg(sc, arg)

    @classmethod
    def run(cls, args):
        self = cls()
        # Some commands need to be run repeatedly on parameter files
        # In fact, this is the rule and the opposite is the exception
        # BUT, we only want to parse the arguments once.
        if cls.npfs > 1:
            self(args)
        else:
            pf_args = getattr(args, "pf", [])
            if len(pf_args) > 1:
                pfs = args.pf
                for pf in pfs:
                    args.pf = pf
                    self(args)
            elif len(pf_args) == 0:
                pfs = []
                self(args)
            else:
                args.pf = getattr(args, 'pf', [None])[0]
                self(args)

class GetParameterFiles(argparse.Action):
    def __call__(self, parser, namespace, values, option_string = None):
        if len(values) == 1:
            pfs = values
        elif len(values) == 2 and namespace.basename is not None:
            pfs = ["%s%04i" % (namespace.basename, r)
                   for r in range(int(values[0]), int(values[1]), namespace.skip) ]
        else:
            pfs = values
        namespace.pf = [_fix_pf(pf) for pf in pfs]

_common_options = dict(
    all     = dict(long="--all", dest="reinstall",
                   default=False, action="store_true",
                   help="Reinstall the full yt stack in the current location."),
    pf      = dict(short="pf", action=GetParameterFiles,
                   nargs="+", help="Parameter files to run on"),
    opf     = dict(action=GetParameterFiles, dest="pf",
                   nargs="*", help="(Optional) Parameter files to run on"),
    axis    = dict(short="-a", long="--axis",
                   action="store", type=int,
                   dest="axis", default=4,
                   help="Axis (4 for all three)"),
    log     = dict(short="-l", long="--log",
                   action="store_true",
                   dest="takelog", default=True,
                   help="Use logarithmic scale for image"),
    linear  = dict(long="--linear",
                   action="store_false",
                   dest="takelog",
                   help="Use linear scale for image"),
    text    = dict(short="-t", long="--text",
                   action="store", type=str,
                   dest="text", default=None,
                   help="Textual annotation"),
    field   = dict(short="-f", long="--field",
                   action="store", type=str,
                   dest="field", default="Density",
                   help="Field to color by"),
    weight  = dict(short="-g", long="--weight",
                   action="store", type=str,
                   dest="weight", default=None,
                   help="Field to weight projections with"),
    cmap    = dict(long="--colormap",
                   action="store", type=str,
                   dest="cmap", default="algae",
                   help="Colormap name"),
    zlim    = dict(short="-z", long="--zlim",
                   action="store", type=float,
                   dest="zlim", default=None,
                   nargs=2,
                   help="Color limits (min, max)"),
    dex     = dict(long="--dex",
                   action="store", type=float,
                   dest="dex", default=None,
                   nargs=1,
                   help="Number of dex above min to display"),
    width   = dict(short="-w", long="--width",
                   action="store", type=float,
                   dest="width", default=None,
                   help="Width in specified units"),
    unit    = dict(short="-u", long="--unit",
                   action="store", type=str,
                   dest="unit", default='1',
                   help="Desired units"),
    center  = dict(short="-c", long="--center",
                   action="store", type=float,
                   dest="center", default=None,
                   nargs=3,
                   help="Center, space separated (-1 -1 -1 for max)"),
    max     = dict(short="-m", long="--max",
                   action="store_true",
                   dest="max",default=False,
                   help="Center the plot on the density maximum"),
    bn      = dict(short="-b", long="--basename",
                   action="store", type=str,
                   dest="basename", default=None,
                   help="Basename of parameter files"),
    output  = dict(short="-o", long="--output",
                   action="store", type=str,
                   dest="output", default="frames/",
                   help="Folder in which to place output images"),
    outputfn= dict(short="-o", long="--output",
                   action="store", type=str,
                   dest="output", default=None,
                   help="File in which to place output"),
    skip    = dict(short="-s", long="--skip",
                   action="store", type=int,
                   dest="skip", default=1,
                   help="Skip factor for outputs"),
    proj    = dict(short="-p", long="--projection",
                   action="store_true",
                   dest="projection", default=False,
                   help="Use a projection rather than a slice"),
    maxw    = dict(long="--max-width",
                   action="store", type=float,
                   dest="max_width", default=1.0,
                   help="Maximum width in code units"),
    minw    = dict(long="--min-width",
                   action="store", type=float,
                   dest="min_width", default=50,
                   help="Minimum width in units of smallest dx (default: 50)"),
    nframes = dict(short="-n", long="--nframes",
                   action="store", type=int,
                   dest="nframes", default=100,
                   help="Number of frames to generate"),
    slabw   = dict(long="--slab-width",
                   action="store", type=float,
                   dest="slab_width", default=1.0,
                   help="Slab width in specified units"),
    slabu   = dict(short="-g", long="--slab-unit",
                   action="store", type=str,
                   dest="slab_unit", default='1',
                   help="Desired units for the slab"),
    ptype   = dict(long="--particle-type",
                   action="store", type=int,
                   dest="ptype", default=2,
                   help="Particle type to select"),
    agecut  = dict(long="--age-cut",
                   action="store", type=float,
                   dest="age_filter", default=None,
                   nargs=2,
                   help="Bounds for the field to select"),
    uboxes  = dict(long="--unit-boxes",
                   action="store_true",
                   dest="unit_boxes",
                   help="Display helpful unit boxes"),
    thresh  = dict(long="--threshold",
                   action="store", type=float,
                   dest="threshold", default=None,
                   help="Density threshold"),
    dm_only = dict(long="--all-particles",
                   action="store_false",
                   dest="dm_only", default=True,
                   help="Use all particles"),
    grids   = dict(long="--show-grids",
                   action="store_true",
                   dest="grids", default=False,
                   help="Show the grid boundaries"),
    time    = dict(long="--time",
                   action="store_true",
                   dest="time", default=False,
                   help="Print time in years on image"),
    contours    = dict(long="--contours",
                   action="store",type=int,
                   dest="contours", default=None,
                   help="Number of Contours for Rendering"),
    contour_width  = dict(long="--contour_width",
                   action="store",type=float,
                   dest="contour_width", default=None,
                   help="Width of gaussians used for rendering."),
    enhance   = dict(long="--enhance",
                   action="store_true",
                   dest="enhance", default=False,
                   help="Enhance!"),
    valrange  = dict(short="-r", long="--range",
                   action="store", type=float,
                   dest="valrange", default=None,
                   nargs=2,
                   help="Range, space separated"),
    up  = dict(long="--up",
                   action="store", type=float,
                   dest="up", default=None,
                   nargs=3,
                   help="Up, space separated"),
    viewpoint  = dict(long="--viewpoint",
                   action="store", type=float,
                   dest="viewpoint", default=[1., 1., 1.],
                   nargs=3,
                   help="Viewpoint, space separated"),
    pixels    = dict(long="--pixels",
                   action="store",type=int,
                   dest="pixels", default=None,
                   help="Number of Pixels for Rendering"),
    halos   = dict(long="--halos",
                   action="store", type=str,
                   dest="halos",default="multiple",
                   help="Run halo profiler on a 'single' halo or 'multiple' halos."),
    halo_radius = dict(long="--halo_radius",
                       action="store", type=float,
                       dest="halo_radius",default=0.1,
                       help="Constant radius for profiling halos if using hop output files with no radius entry. Default: 0.1."),
    halo_radius_units = dict(long="--halo_radius_units",
                             action="store", type=str,
                             dest="halo_radius_units",default="1",
                             help="Units for radius used with --halo_radius flag. Default: '1' (code units)."),
    halo_hop_style = dict(long="--halo_hop_style",
                          action="store", type=str,
                          dest="halo_hop_style",default="new",
                          help="Style of hop output file.  'new' for yt_hop files and 'old' for enzo_hop files."),
    halo_parameter_file = dict(long="--halo_parameter_file",
                               action="store", type=str,
                               dest="halo_parameter_file",default=None,
                               help="HaloProfiler parameter file."),
    make_profiles = dict(long="--make_profiles",
                         action="store_true", default=False,
                         help="Make profiles with halo profiler."),
    make_projections = dict(long="--make_projections",
                            action="store_true", default=False,
                            help="Make projections with halo profiler.")

    )

def _get_yt_stack_date():
    if "YT_DEST" not in os.environ:
        print "Could not determine when yt stack was last updated."
        return
    date_file = os.path.join(os.environ["YT_DEST"], ".yt_update")
    if not os.path.exists(date_file):
        print "Could not determine when yt stack was last updated."
        return
    print "".join(file(date_file, 'r').readlines())
    print "To update all dependencies, run \"yt update --all\"."

def _update_yt_stack(path):
    "Rerun the install script to updated all dependencies."

    install_script = os.path.join(path, "doc/install_script.sh")
    if not os.path.exists(install_script):
        print
        print "Install script not found!"
        print "The install script should be here: %s," % install_script
        print "but it was not."
        return

    print
    print "We will now attempt to update the yt stack located at:"
    print "    %s." % os.environ["YT_DEST"]
    print
    print "[hit enter to continue or Ctrl-C to stop]"
    try:
        raw_input()
    except:
        sys.exit(0)
    os.environ["REINST_YT"] = "1"
    ret = subprocess.call(["bash", install_script])
    print
    if ret:
        print "The install script seems to have failed."
        print "Check the output above."
    else:
        print "The yt stack has been updated successfully."
        print "Now get back to work!"

def _update_hg(path, skip_rebuild = False):
    from mercurial import hg, ui, commands
    f = open(os.path.join(path, "yt_updater.log"), "a")
    u = ui.ui()
    u.pushbuffer()
    config_fn = os.path.join(path, ".hg", "hgrc")
    print "Reading configuration from ", config_fn
    u.readconfig(config_fn)
    repo = hg.repository(u, path)
    commands.pull(u, repo)
    f.write(u.popbuffer())
    f.write("\n\n")
    u.pushbuffer()
    commands.identify(u, repo)
    if "+" in u.popbuffer():
        print "Can't rebuild modules by myself."
        print "You will have to do this yourself.  Here's a sample commands:"
        print
        print "    $ cd %s" % (path)
        print "    $ hg up"
        print "    $ %s setup.py develop" % (sys.executable)
        return 1
    print "Updating the repository"
    f.write("Updating the repository\n\n")
    commands.update(u, repo, check=True)
    if skip_rebuild: return
    f.write("Rebuilding modules\n\n")
    p = subprocess.Popen([sys.executable, "setup.py", "build_ext", "-i"], cwd=path,
                        stdout = subprocess.PIPE, stderr = subprocess.STDOUT)
    stdout, stderr = p.communicate()
    f.write(stdout)
    f.write("\n\n")
    if p.returncode:
        print "BROKEN: See %s" % (os.path.join(path, "yt_updater.log"))
        sys.exit(1)
    f.write("Successful!\n")
    print "Updated successfully."

def _get_hg_version(path):
    from mercurial import hg, ui, commands
    u = ui.ui()
    u.pushbuffer()
    repo = hg.repository(u, path)
    commands.identify(u, repo)
    return u.popbuffer()

def get_yt_version():
    try:
        from yt.__hg_version__ import hg_version
        return hg_version
    except ImportError:
        pass
    import pkg_resources
    yt_provider = pkg_resources.get_provider("yt")
    path = os.path.dirname(yt_provider.module_path)
    if not os.path.isdir(os.path.join(path, ".hg")): return None
    version = _get_hg_version(path)[:12]
    return version

# This code snippet is modified from Georg Brandl
def bb_apicall(endpoint, data, use_pass = True):
    uri = 'https://api.bitbucket.org/1.0/%s/' % endpoint
    # since bitbucket doesn't return the required WWW-Authenticate header when
    # making a request without Authorization, we cannot use the standard urllib2
    # auth handlers; we have to add the requisite header from the start
    if data is not None:
        data = urllib.urlencode(data)
    req = urllib2.Request(uri, data)
    if use_pass:
        username = raw_input("Bitbucket Username? ")
        password = getpass.getpass()
        upw = '%s:%s' % (username, password)
        req.add_header('Authorization', 'Basic %s' % base64.b64encode(upw).strip())
    return urllib2.urlopen(req).read()

def _get_yt_supp(uu):
    supp_path = os.path.join(os.environ["YT_DEST"], "src",
                             "yt-supplemental")
    # Now we check that the supplemental repository is checked out.
    from mercurial import hg, ui, commands
    if not os.path.isdir(supp_path):
        print
        print "*** The yt-supplemental repository is not checked ***"
        print "*** out.  I can do this for you, but because this ***"
        print "*** is a delicate act, I require you to respond   ***"
        print "*** to the prompt with the word 'yes'.            ***"
        print
        response = raw_input("Do you want me to try to check it out? ")
        if response != "yes":
            print
            print "Okay, I understand.  You can check it out yourself."
            print "This command will do it:"
            print
            print "$ hg clone http://hg.yt-project.org/yt-supplemental/ ",
            print "%s" % (supp_path)
            print
            sys.exit(1)
        rv = commands.clone(uu,
                "http://hg.yt-project.org/yt-supplemental/", supp_path)
        if rv:
            print "Something has gone wrong.  Quitting."
            sys.exit(1)
    # Now we think we have our supplemental repository.
    return supp_path

class YTBootstrapDevCmd(YTCommand):
    name = "bootstrap_dev"
    description = \
        """
        Bootstrap a yt development environment
        """
    def __call__(self, args):
        from mercurial import hg, ui, commands
        import imp
        import getpass
        import json
        uu = ui.ui()
        print
        print "Hi there!  Welcome to the yt development bootstrap tool."
        print
        print "This should get you started with mercurial as well as a few"
        print "other handy things"
        print
        # We have to do a couple things.
        # First, we check that YT_DEST is set.
        if "YT_DEST" not in os.environ:
            print
            print "*** You must set the environment variable YT_DEST ***"
            print "*** to point to the installation location!        ***"
            print
            sys.exit(1)
        supp_path = _get_yt_supp(uu)
        print
        print "I have found the yt-supplemental repository at %s" % (supp_path)
        print
        print "Let's load up and check what we need to do to get up and"
        print "running."
        print
        print "There are three stages:"
        print
        print " 1. Setting up your ~/.hgrc to have a username."
        print " 2. Setting up your bitbucket user account and the hgbb"
        print "    extension."
        print
        firstname = lastname = email_address = bbusername = repo_list = None
        # Now we try to import the cedit extension.
        try:
            result = imp.find_module("cedit", [supp_path])
        except ImportError:
            print "I was unable to find the 'cedit' module in %s" % (supp_path)
            print "This may be due to a broken checkout."
            print "Sorry, but I'm going to bail."
            sys.exit(1)
        cedit = imp.load_module("cedit", *result)
        try:
            result = imp.find_module("hgbb", [supp_path + "/hgbb"])
        except ImportError:
            print "I was unable to find the 'hgbb' module in %s" % (supp_path)
            print "This may be due to a broken checkout."
            print "Sorry, but I'm going to bail."
            sys.exit(1)
        hgbb = imp.load_module("hgbb", *result)
        if uu.config("ui","username",None) is None:
            print "You don't have a username specified in your ~/.hgrc."
            print "Let's set this up.  If you would like to quit at any time,"
            print "hit Ctrl-C."
            print
            firstname = raw_input("What is your first name? ")
            lastname = raw_input("What is your last name? ")
            email_address = raw_input("What is your email address? ")
            print
            print "Thanks.  I will now add a username of this form to your"
            print "~/.hgrc file:"
            print
            print "    %s %s <%s>" % (firstname, lastname, email_address)
            print
            loki = raw_input("Press enter to go on, Ctrl-C to exit.")
            print
            cedit.setuser(uu, name="%s %s" % (firstname, lastname),
                              email="%s" % (email_address),
                              local=False, username=False)
            print
        else:
            print "Looks like you already have a username!"
            print "We'll skip that step, then."
            print
        print "Now we'll set up BitBucket user.  If you would like to do this"
        print "yourself, please visit:"
        print " https://bitbucket.org/account/signup/?plan=5_users"
        print "for a free account."
        print
        loki = raw_input("Do you have a BitBucket.org user already? [yes/no]")
        if loki.strip().upper() == "YES":
            bbusername = raw_input("Okay, cool.  What is your username?  ").strip()
            # Now we get information about the username.
            if firstname is None or lastname is None:
                rv = hgbb._bb_apicall(uu, "users/%s" % bbusername, None, False)
                rv = json.loads(rv)
                firstname = rv['user']["first_name"]
                lastname = rv['user']["last_name"]
                repo_list = rv['repositories']
                print "Retrieved your info:"
                print "  username:   %s" % (bbusername)
                print "  first name: %s" % (firstname)
                print "  last name:  %s" % (lastname)
        elif loki.strip().upper() == "NO":
            print "Okay, we can set you up with one.  It's probably better for"
            print "it to be all lowercase letter."
            print
            bbusername = raw_input("What is your desired username? ").strip()
            if firstname is None:
                firstname = raw_input("What's your first name? ").strip()
            if lastname is None:
                lastname = raw_input("What's your last name? ").strip()
            if email_address is None:
                email_address = raw_input("What's your email address? ").strip()
            print
            print "Okay, I'll see if I can create a user with this information:"
            print "  username:   %s" % (bbusername)
            print "  first name: %s" % (firstname)
            print "  last name:  %s" % (lastname)
            print "  email:      %s" % (email_address)
            print
            print "Now, I'm going to ask for a password.  This password will"
            print "be transmitted over HTTPS (not HTTP) and will not be stored"
            print "in any local file.  But, it will be stored in memory for"
            print "the duration of the user-creation process."
            print
            while 1:
                password1 = getpass.getpass("Password? ")
                password2 = getpass.getpass("Confirm? ")
                if password1 == password2: break
                print "Sorry, they didn't match!  Let's try again."
                print
            rv = hgbb._bb_apicall(uu, "newuser",
                                    dict(username=bbusername,
                                         password=password1,
                                         email=email_address,
                                         first_name = firstname,
                                         last_name = lastname),
                                   False)
            del password1, password2
            if str(json.loads(rv)['username']) == bbusername:
                print "Successful!  You probably just got an email asking you"
                print "to confirm this."
            else:
                print "Okay, something is wrong.  Quitting!"
                sys.exit(1)
        else:
            print "Not really sure what you replied with.  Quitting!"
            sys.exit(1)
        # We're now going to do some modification of the hgrc.
        # We need an hgrc first.
        hgrc_path = [cedit.config.defaultpath("user", uu)]
        hgrc_path = cedit.config.verifypaths(hgrc_path)
        # Now we set up the hgbb extension
        if uu.config("extensions","config",None) is None:
            # cedit is a module, but hgbb is a file.  So we call dirname here.
            cedit_path = os.path.dirname(cedit.__file__)
            print "Now we're going to turn on the cedit extension in:"
            print "    ", hgrc_path
            print "This will enable you to edit your configuration from the"
            print "command line.  Mainly, this is useful to use the command"
            print "'addsource', which will let you add a new source to a given"
            print "mercurial repository -- like a fork, or your own fork."
            print
            print "This constitutes adding the path to the cedit extension,"
            print "which will look like this:"
            print
            print "   [extensions]"
            print "   config=%s" % cedit_path
            print
            loki = raw_input("Press enter to go on, Ctrl-C to exit.")
            cedit.config.setoption(uu, hgrc_path, "extensions.config=%s" % cedit_path)
        if uu.config("extensions","hgbb",None) is None:
            hgbb_path = hgbb.__file__
            if hgbb_path.endswith(".pyc"): hgbb_path = hgbb_path[:-1]
            print "Now we're going to turn on the hgbb extension in:"
            print "    ", hgrc_path
            print "This will enable you to access BitBucket more easily, as well"
            print "as create repositories from the command line."
            print
            print "This constitutes adding the path to the hgbb extension,"
            print "which will look like this:"
            print
            print "   [extensions]"
            print "   hgbb=%s" % hgbb_path
            print
            loki = raw_input("Press enter to go on, Ctrl-C to exit.")
            cedit.config.setoption(uu, hgrc_path, "extensions.hgbb=%s" % hgbb_path)
        if uu.config("bb","username", None) is None:
            print "We'll now set up your username for BitBucket."
            print "We will add this:"
            print
            print "   [bb]"
            print "   username = %s" % (bbusername)
            print
            loki = raw_input("Press enter to go on, Ctrl-C to exit.")
            cedit.config.setoption(uu, hgrc_path, "bb.username=%s" % bbusername)
        # We now reload the UI's config file so that it catches the [bb]
        # section changes.
        uu.readconfig(hgrc_path[0])
        try:
            import lxml
            install_lxml = False
        except ImportError:
            install_lxml = True
        if install_lxml:
            print "You are missing the lxml package.  Installing."
            import pip
            rv = pip.main(["install", "lxml"])
            if rv == 1:
                print "Unable to install lxml.  Please report this bug to yt-users."
                sys.exit(1)
        print
        print "All done!"
        print
        print "You're now set up to develop using Mercurial and BitBucket."
        print
        print "Good luck!"

class YTBugreportCmd(YTCommand):
    name = "bugreport"
    description = \
        """
        Report a bug in yt

        """

    def __call__(self, args):
        print "==============================================================="
        print
        print "Hi there!  Welcome to the yt bugreport taker."
        print
        print "==============================================================="
        print "At any time in advance of the upload of the bug, you should feel free"
        print "to ctrl-C out and submit the bug report manually by going here:"
        print "   http://hg.yt-project.org/yt/issues/new"
        print
        print "Also, in order to submit a bug through this interface, you"
        print "need a Bitbucket account. If you don't have one, exit this "
        print "bugreport now and run the 'yt bootstrap_dev' command to create one."
        print
        print "Have you checked the existing bug reports to make"
        print "sure your bug has not already been recorded by someone else?"
        print "   http://hg.yt-project.org/yt/issues?status=new&status=open"
        print
        print "Finally, are you sure that your bug is, in fact, a bug? It might"
        print "simply be a misunderstanding that could be cleared up by"
        print "visiting the yt irc channel or getting advice on the email list:"
        print "   http://yt-project.org/irc.html"
        print "   http://lists.spacepope.org/listinfo.cgi/yt-users-spacepope.org"
        print
        summary = raw_input("Press <enter> if you remain firm in your conviction to continue.")
        print
        print
        print "Okay, sorry about that. How about a nice, pithy ( < 12 words )"
        print "summary of the bug?  (e.g. 'Particle overlay problem with parallel "
        print "projections')"
        print
        try:
            current_version = get_yt_version()
        except:
            current_version = "Unavailable"
        summary = raw_input("Summary? ")
        bugtype = "bug"
        data = dict(title = summary, type=bugtype)
        print
        print "Okay, now let's get a bit more information."
        print
        print "Remember that if you want to submit a traceback, you can run"
        print "any script with --paste or --detailed-paste to submit it to"
        print "the pastebin and then include the link in this bugreport."
        if "EDITOR" in os.environ:
            print
            print "Press enter to spawn your editor, %s" % os.environ["EDITOR"]
            loki = raw_input()
            tf = tempfile.NamedTemporaryFile(delete=False)
            fn = tf.name
            tf.close()
            popen = subprocess.call("$EDITOR %s" % fn, shell = True)
            content = open(fn).read()
            try:
                os.unlink(fn)
            except:
                pass
        else:
            print
            print "Couldn't find an $EDITOR variable.  So, let's just take"
            print "take input here.  Type up your summary until you're ready"
            print "to be done, and to signal you're done, type --- by itself"
            print "on a line to signal your completion."
            print
            print "(okay, type now)"
            print
            lines = []
            while 1:
                line = raw_input()
                if line.strip() == "---": break
                lines.append(line)
            content = "\n".join(lines)
        content = "Reporting Version: %s\n\n%s" % (current_version, content)
        endpoint = "repositories/yt_analysis/yt/issues"
        data['content'] = content
        print
        print "==============================================================="
        print
        print "Okay, we're going to submit with this:"
        print
        print "Summary: %s" % (data['title'])
        print
        print "---"
        print content
        print "---"
        print
        print "==============================================================="
        print
        print "Is that okay?  If not, hit ctrl-c.  Otherwise, enter means"
        print "'submit'.  Next we'll ask for your Bitbucket Username."
        print "If you don't have one, run the 'yt bootstrap_dev' command."
        print
        loki = raw_input()
        retval = bb_apicall(endpoint, data, use_pass=True)
        import json
        retval = json.loads(retval)
        url = "http://hg.yt-project.org/yt/issue/%s" % retval['local_id']
        print
        print "==============================================================="
        print
        print "Thanks for your bug report!  Together we'll make yt totally bug free!"
        print "You can view bug report here:"
        print "   %s" % url
        print
        print "Keep in touch!"
        print

class YTHopCmd(YTCommand):
    args = ('outputfn','bn','thresh','dm_only','skip', 'pf')
    name = "hop"
    description = \
        """
        Run HOP on one or more datasets

        """

    def __call__(self, args):
        pf = args.pf
        kwargs = {'dm_only' : args.dm_only}
        if args.threshold is not None: kwargs['threshold'] = args.threshold
        hop_list = HaloFinder(pf, **kwargs)
        if args.output is None: fn = "%s.hop" % pf
        else: fn = args.output
        hop_list.write_out(fn)

class YTHubRegisterCmd(YTCommand):
    name = "hub_register"
    description = \
        """
        Register a user on the Hub: http://hub.yt-project.org/
        """
    def __call__(self, args):
        # We need these pieces of information:
        #   1. Name
        #   2. Email
        #   3. Username
        #   4. Password (and password2)
        #   5. (optional) URL
        #   6. "Secret" key to make it epsilon harder for spammers
        if ytcfg.get("yt","hub_api_key") != "":
            print "You seem to already have an API key for the hub in"
            print "~/.yt/config .  Delete this if you want to force a"
            print "new user registration."
        print "Awesome!  Let's start by registering a new user for you."
        print "Here's the URL, for reference: http://hub.yt-project.org/ "
        print
        print "As always, bail out with Ctrl-C at any time."
        print
        print "What username would you like to go by?"
        print
        username = raw_input("Username? ")
        if len(username) == 0: sys.exit(1)
        print
        print "To start out, what's your name?"
        print
        name = raw_input("Name? ")
        if len(name) == 0: sys.exit(1)
        print
        print "And your email address?"
        print
        email = raw_input("Email? ")
        if len(email) == 0: sys.exit(1)
        print
        print "Please choose a password:"
        print
        while 1:
            password1 = getpass.getpass("Password? ")
            password2 = getpass.getpass("Confirm? ")
            if len(password1) == 0: continue
            if password1 == password2: break
            print "Sorry, they didn't match!  Let's try again."
            print
        print
        print "Would you like a URL displayed for your user?"
        print "Leave blank if no."
        print
        url = raw_input("URL? ")
        print
        print "Okay, press enter to register.  You should receive a welcome"
        print "message at %s when this is complete." % email
        print
        loki = raw_input()
        data = dict(name = name, email = email, username = username,
                    password = password1, password2 = password2,
                    url = url, zap = "rowsdower")
        data = urllib.urlencode(data)
        hub_url = "https://hub.yt-project.org/create_user"
        req = urllib2.Request(hub_url, data)
        try:
            status = urllib2.urlopen(req).read()
        except urllib2.HTTPError as exc:
            if exc.code == 400:
                print "Sorry, the Hub couldn't create your user."
                print "You can't register duplicate users, which is the most"
                print "common cause of this error.  All values for username,"
                print "name, and email must be unique in our system."
                sys.exit(1)
        except urllib2.URLError as exc:
            print "Something has gone wrong.  Here's the error message."
            raise exc
        print
        print "SUCCESS!"
        print


class YTHubSubmitCmd(YTCommand):
    name = "hub_submit"
    args = (
            dict(long="--repo", action="store", type=str,
                 dest="repo", default=".", help="Repository to upload"),
           )
    description = \
        """
        Submit a mercurial repository to the yt Hub
        (http://hub.yt-project.org/), creating a BitBucket repo in the process
        if necessary.
        """

    def __call__(self, args):
        import imp
        api_key = ytcfg.get("yt","hub_api_key")
        url = ytcfg.get("yt","hub_url")
        if api_key == '':
            print
            print "You must create an API key before uploading."
            print "https://data.yt-project.org/getting_started.html"
            print
            sys.exit(1)
        from mercurial import hg, ui, commands, error, config
        uri = "http://hub.yt-project.org/3rdparty/API/api.php"
        uu = ui.ui()
        supp_path = _get_yt_supp(uu)
        try:
            result = imp.find_module("cedit", [supp_path])
        except ImportError:
            print "I was unable to find the 'cedit' module in %s" % (supp_path)
            print "This may be due to a broken checkout."
            print "Sorry, but I'm going to bail."
            sys.exit(1)
        cedit = imp.load_module("cedit", *result)
        try:
            result = imp.find_module("hgbb", [supp_path + "/hgbb"])
        except ImportError:
            print "I was unable to find the 'hgbb' module in %s" % (supp_path)
            print "This may be due to a broken checkout."
            print "Sorry, but I'm going to bail."
            sys.exit(1)
        hgbb = imp.load_module("hgbb", *result)
        try:
            repo = hg.repository(uu, args.repo)
            conf = config.config()
            if os.path.exists(os.path.join(args.repo,".hg","hgrc")):
                conf.read(os.path.join(args.repo, ".hg", "hgrc"))
            needs_bb = True
            if "paths" in conf.sections():
                default = conf['paths'].get("default", "")
                if default.startswith("bb://") or "bitbucket.org" in default:
                    needs_bb = False
                    bb_url = default
                else:
                    for alias, value in conf["paths"].items():
                        if value.startswith("bb://") or "bitbucket.org" in value:
                            needs_bb = False
                            bb_url = value
                            break
        except error.RepoError:
            print "Unable to find repo at:"
            print "   %s" % (os.path.abspath(args.repo))
            print
            print "Would you like to initialize one?  If this message"
            print "surprises you, you should perhaps press Ctrl-C to quit."
            print "Otherwise, type 'yes' at the prompt."
            print
            loki = raw_input("Create repo? ")
            if loki.upper().strip() != "YES":
                print "Okay, rad -- we'll let you handle it and get back to",
                print " us."
                return 1
            commands.init(uu, dest=args.repo)
            repo = hg.repository(uu, args.repo)
            commands.add(uu, repo)
            commands.commit(uu, repo, message="Initial automated import by yt")
            needs_bb = True
        if needs_bb:
            print
            print "Your repository is not yet on BitBucket, as near as I can tell."
            print "Would you like to create a repository there and upload to it?"
            print "Without this, I don't know what URL to submit!"
            print
            print "Type 'yes' to accept."
            print
            loki = raw_input("Upload to BitBucket? ")
            if loki.upper().strip() != "YES": return 1
            hgrc_path = [cedit.config.defaultpath("user", uu)]
            hgrc_path = cedit.config.verifypaths(hgrc_path)
            uu.readconfig(hgrc_path[0])
            bb_username = uu.config("bb", "username", None)
            if bb_username is None:
                print "Can't find your Bitbucket username.  Run the command:"
                print
                print "$ yt bootstrap_dev"
                print
                print "to get set up and ready to go."
                return 1
            bb_repo_name = os.path.basename(os.path.abspath(args.repo))
            print
            print "I am now going to create the repository:"
            print "    ", bb_repo_name
            print "on BitBucket.org and upload this repository to that."
            print "If that is not okay, please press Ctrl-C to quit."
            print
            loki = raw_input("Press Enter to continue.")
            data = dict(name=bb_repo_name)
            hgbb._bb_apicall(uu, 'repositories', data)
            print
            print "Created repository!  Now I will set this as the default path."
            print
            bb_url = "https://%s@bitbucket.org/%s/%s" % (
                        bb_username, bb_username, bb_repo_name)
            cedit.config.addsource(uu, repo, "default", bb_url)
            commands.push(uu, repo, bb_url)
            # Now we reset
            bb_url = "https://bitbucket.org/%s/%s" % (
                        bb_username, bb_repo_name)
        if bb_url.startswith("bb://"):
            bb_username, bb_repo_name = bb_url.split("/")[-2:]
            bb_url = "https://bitbucket.org/%s/%s" % (
                bb_username, bb_repo_name)
        # Now we can submit
        print
        print "Okay.  Now we're ready to submit to the Hub."
        print "Remember, you can go to the Hub at any time at"
        print " http://hub.yt-project.org/"
        print
        print "(Especially if you don't have a user yet!  We can wait.)"
        print

        categories = {
            1: "News",
            2: "Documents",
            3: "Simulation Management",
            4: "Data Management",
            5: "Analysis and Visualization",
            6: "Paper Repositories",
            7: "Astrophysical Utilities",
            8: "yt Scripts"
        }
        cat_id = -1
        while cat_id not in categories:
            print
            for i, n in sorted(categories.items()):
                print "%i. %s" % (i, n)
            print
            cat_id = int(raw_input("Which category number does your script fit into? "))
        print
        print "What is the title of your submission? (Usually a repository name) "
        title = raw_input("Title? ")
        print
        print "Give us a very brief summary of the project -- enough to get someone"
        print "interested enough to click the link and see what it's about.  This"
        print "should be a few sentences at most."
        print
        summary = raw_input("Summary? ")
        print
        print "Is there a URL that you'd like to point the image to?  Just hit"
        print "enter if no."
        print
        image_url = raw_input("Image URL? ").strip()
        print
        print "Okay, we're going to submit!  Press enter to submit, Ctrl-C to back out."
        print
        loki = raw_input()

        mpd = MinimalProjectDescription(title, bb_url, summary,
                categories[cat_id], image_url)
        mpd.upload()

class YTInstInfoCmd(YTCommand):
    name = "instinfo"
    args = (
            dict(short="-u", long="--update-source", action="store_true",
                 default = False,
                 help="Update the yt installation, if able"),
            dict(short="-o", long="--output-version", action="store",
                  default = None, dest="outputfile",
                  help="File into which the current revision number will be" +
                       "stored")
           )
    description = \
        """
        Get some information about the yt installation

        """

    def __call__(self, opts):
        import pkg_resources
        yt_provider = pkg_resources.get_provider("yt")
        path = os.path.dirname(yt_provider.module_path)
        print
        print "yt module located at:"
        print "    %s" % (path)
        update_supp = False
        if "YT_DEST" in os.environ:
            spath = os.path.join(
                     os.environ["YT_DEST"], "src", "yt-supplemental")
            if os.path.isdir(spath):
                print "The supplemental repositories are located at:"
                print "    %s" % (spath)
                update_supp = True
        vstring = get_yt_version()
        if vstring is not None:
            print
            print "The current version of the code is:"
            print
            print "---"
            print vstring.strip()
            print "---"
            print
            if "site-packages" not in path:
                print "This installation CAN be automatically updated."
                if opts.update_source:
                    update_hg(path)
                    print "Updated successfully."
                _get_yt_stack_date()
        elif opts.update_source:
            print
            print "YT site-packages not in path, so you must"
            print "update this installation manually by committing and"
            print "merging your modifications to the code before"
            print "updating to the newest changeset."
            print
        if vstring is not None and opts.outputfile is not None:
            open(opts.outputfile, "w").write(vstring)

class YTLoadCmd(YTCommand):
    name = "load"
    description = \
        """
        Load a single dataset into an IPython instance

        """

    args = ("pf", )

    def __call__(self, args):
        if args.pf is None:
            print "Could not load file."
            sys.exit()
        import yt.mods

        import IPython
        from distutils import version
        if version.LooseVersion(IPython.__version__) <= version.LooseVersion('0.10'):
            api_version = '0.10'
        else:
            api_version = '0.11'

        local_ns = yt.mods.__dict__.copy()
        local_ns['pf'] = args.pf

        if api_version == '0.10':
            shell = IPython.Shell.IPShellEmbed()
            shell(local_ns = local_ns,
                  header =
                  "\nHi there!  Welcome to yt.\n\nWe've loaded your parameter file as 'pf'.  Enjoy!"
                  )
        else:
            from IPython.config.loader import Config
            import sys
            cfg = Config()
            # prepend sys.path with current working directory
            sys.path.insert(0,'')
            IPython.embed(config=cfg,user_ns=local_ns)


class YTMapserverCmd(YTCommand):
    args = ("proj", "field", "weight",
            dict(short="-a", long="--axis", action="store", type=int,
                 dest="axis", default=0, help="Axis (4 for all three)"),
            dict(short ="-o", long="--host", action="store", type=str,
                   dest="host", default=None, help="IP Address to bind on"),
            "pf",
            )

    name = "mapserver"
    description = \
        """
        Serve a plot in a GMaps-style interface

        """

    def __call__(self, args):
        pf = args.pf
        pc=PlotCollection(pf, center=0.5*(pf.domain_left_edge +
                                          pf.domain_right_edge))
        if args.axis == 4:
            print "Doesn't work with multiple axes!"
            return
        if args.projection:
            p = pc.add_projection(args.field, args.axis, weight_field=args.weight)
        else:
            p = pc.add_slice(args.field, args.axis)
        from yt.gui.reason.pannable_map import PannableMapServer
        mapper = PannableMapServer(p.data, args.field)
        import yt.utilities.bottle as bottle
        bottle.debug(True)
        if args.host is not None:
            colonpl = args.host.find(":")
            if colonpl >= 0:
                port = int(args.host.split(":")[-1])
                args.host = args.host[:colonpl]
            else:
                port = 8080
            bottle.run(server='rocket', host=args.host, port=port)
        else:
            bottle.run(server='rocket')

class YTPastebinCmd(YTCommand):
    name = "pastebin"
    args = (
             dict(short="-l", long="--language", action="store",
                  default = None, dest="language",
                  help="Use syntax highlighter for the file in language"),
             dict(short="-L", long="--languages", action="store_true",
                  default = False, dest="languages",
                  help="Retrive a list of supported languages"),
             dict(short="-e", long="--encoding", action="store",
                  default = 'utf-8', dest="encoding",
                  help="Specify the encoding of a file (default is "
                        "utf-8 or guessing if available)"),
             dict(short="-b", long="--open-browser", action="store_true",
                  default = False, dest="open_browser",
                  help="Open the paste in a web browser"),
             dict(short="-p", long="--private", action="store_true",
                  default = False, dest="private",
                  help="Paste as private"),
             dict(short="-c", long="--clipboard", action="store_true",
                  default = False, dest="clipboard",
                  help="File to output to; else, print."),
             dict(short="file", type=str),
            )
    description = \
        """
        Post a script to an anonymous pastebin

        """

    def __call__(self, args):
        import yt.utilities.lodgeit as lo
        lo.main(args.file, languages=args.languages, language=args.language,
                 encoding=args.encoding, open_browser=args.open_browser,
                 private=args.private, clipboard=args.clipboard)

class YTPastebinGrabCmd(YTCommand):
    args = (dict(short="number", type=str),)
    name = "pastebin_grab"
    description = \
        """
        Print an online pastebin to STDOUT for local use.
        """

    def __call__(self, args):
        import yt.utilities.lodgeit as lo
        lo.main( None, download=args.number )

class YTNotebookUploadCmd(YTCommand):
    args = (dict(short="file", type=str),)
    description = \
        """
        Upload an IPython notebook to hub.yt-project.org.
        """

    name = "upload_notebook"
    def __call__(self, args):
        filename = args.file
        if not os.path.isfile(filename):
            raise IOError(filename)
        if not filename.endswith(".ipynb"):
            print "File must be an IPython notebook!"
            return 1
        import json
        try:
            t = json.loads(open(filename).read())['metadata']['name']
        except (ValueError, KeyError):
            print "File does not appear to be an IPython notebook."
        from yt.utilities.minimal_representation import MinimalNotebook
        mn = MinimalNotebook(filename, t)
        rv = mn.upload()
        print "Upload successful!"
        print
        print "To access your raw notebook go here:"
        print
        print "  %s" % (rv['url'])
        print
        print "To view your notebook go here:"
        print
        print "  %s" % (rv['url'].replace("/go/", "/nb/"))
        print

class YTPlotCmd(YTCommand):
    args = ("width", "unit", "bn", "proj", "center", "zlim", "axis", "field",
            "weight", "skip", "cmap", "output", "grids", "time", "pf", "max",
            "log", "linear")

    name = "plot"

    description = \
        """
        Create a set of images

        """

    def __call__(self, args):
        pf = args.pf
        center = args.center
        if args.center == (-1,-1,-1):
            mylog.info("No center fed in; seeking.")
            v, center = pf.h.find_max("Density")
        if args.max:
            v, center = pf.h.find_max("Density")
        elif args.center is None:
            center = 0.5*(pf.domain_left_edge + pf.domain_right_edge)
        center = np.array(center)
        if pf.dimensionality < 3:
            dummy_dimensions = np.nonzero(pf.h.grids[0].ActiveDimensions <= 1)
            axes = ensure_list(dummy_dimensions[0][0])
        elif args.axis == 4:
            axes = range(3)
        else:
            axes = [args.axis]

        unit = args.unit
        if unit is None:
            unit = 'unitary'
        if args.width is None:
            width = None
        else:
            width = (args.width, args.unit)

        for ax in axes:
            mylog.info("Adding plot for axis %i", ax)
            if args.projection:
                plt = ProjectionPlot(pf, ax, args.field, center=center,
                                     width=width,
                                     weight_field=args.weight)
            else:
                plt = SlicePlot(pf, ax, args.field, center=center,
                                width=width)
            if args.grids:
                plt.annotate_grids()
            if args.time:
                time = pf.current_time*pf['Time']*pf['years']
                plt.annotate_text((0.2,0.8), 't = %5.2e yr'%time)

            plt.set_cmap(args.field, args.cmap)
            plt.set_log(args.field, args.takelog)
            if args.zlim:
                plt.set_zlim(args.field,*args.zlim)
            ensure_dir_exists(args.output)
            plt.save(os.path.join(args.output,"%s" % (pf)))

class YTRenderCmd(YTCommand):

    args = ("width", "unit", "center","enhance",'outputfn',
            "field", "cmap", "contours", "viewpoint", "linear",
            "pixels", "up", "valrange", "log","contour_width", "pf")
    name = "render"
    description = \
        """
        Create a simple volume rendering
        """

    def __call__(self, args):
        pf = args.pf
        center = args.center
        if args.center == (-1,-1,-1):
            mylog.info("No center fed in; seeking.")
            v, center = pf.h.find_max("Density")
        elif args.center is None:
            center = 0.5*(pf.domain_left_edge + pf.domain_right_edge)
        center = np.array(center)

        L = args.viewpoint
        if L is None:
            L = [1.]*3
        L = np.array(args.viewpoint)

        unit = args.unit
        if unit is None:
            unit = '1'
        width = args.width
        if width is None:
            width = 0.5*(pf.domain_right_edge - pf.domain_left_edge)
        width /= pf[unit]

        N = args.pixels
        if N is None:
            N = 512

        up = args.up
        if up is None:
            up = [0.,0.,1.]

        field = args.field
        if field is None:
            field = 'Density'

        log = args.takelog
        if log is None:
            log = True

        myrange = args.valrange
        if myrange is None:
            roi = pf.h.region(center, center-width, center+width)
            mi, ma = roi.quantities['Extrema'](field)[0]
            if log:
                mi, ma = np.log10(mi), np.log10(ma)
        else:
            mi, ma = myrange[0], myrange[1]

        n_contours = args.contours
        if n_contours is None:
            n_contours = 7

        contour_width = args.contour_width

        cmap = args.cmap
        if cmap is None:
            cmap = 'jet'
        tf = ColorTransferFunction((mi-2, ma+2))
        tf.add_layers(n_contours,w=contour_width,col_bounds = (mi,ma), colormap=cmap)

        cam = pf.h.camera(center, L, width, (N,N), transfer_function=tf)
        image = cam.snapshot()

        if args.enhance:
            for i in range(3):
                image[:,:,i] = image[:,:,i]/(image[:,:,i].mean() + 5.*image[:,:,i].std())
            image[image>1.0]=1.0

        save_name = args.output
        if save_name is None:
            save_name = "%s"%pf+"_"+field+"_rendering.png"
        if not '.png' in save_name:
            save_name += '.png'
        if cam.comm.rank != -1:
            write_bitmap(image,save_name)

class YTRPDBCmd(YTCommand):
    name = "rpdb"
    description = \
        """
        Connect to a currently running (on localhost) rpd session.

        Commands run with --rpdb will trigger an rpdb session with any
        uncaught exceptions.

        """
    args = (
            dict(short="-t", long="--task", action="store",
                 default = 0, dest='task',
                 help="Open a web browser."),
           )

    def __call__(self, args):
        import rpdb
        rpdb.run_rpdb(int(args.task))

class YTNotebookCmd(YTCommand):
    name = ["notebook"]
    args = (
            dict(short="-o", long="--open-browser", action="store_true",
                 default = False, dest='open_browser',
                 help="Open a web browser."),
            dict(short="-p", long="--port", action="store",
                 default = 0, dest='port',
                 help="Port to listen on; defaults to auto-detection."),
            dict(short="-n", long="--no-password", action="store_true",
                 default = False, dest='no_password',
                 help="If set, do not prompt or use a password."),
            )
    description = \
        """
        Run the IPython Notebook
        """
    def __call__(self, args):
        kwargs = {}
        from IPython.frontend.html.notebook.notebookapp import NotebookApp
        pw = ytcfg.get("yt", "notebook_password")
        if len(pw) == 0 and not args.no_password:
            import IPython.lib
            pw = IPython.lib.passwd()
            print "If you would like to use this password in the future,"
            print "place a line like this inside the [yt] section in your"
            print "yt configuration file at ~/.yt/config"
            print
            print "notebook_password = %s" % pw
            print
        elif args.no_password:
            pw = None
        if args.port != 0:
            kwargs['port'] = int(args.port)
        if pw is not None:
            kwargs['password'] = pw
        app = NotebookApp(open_browser=args.open_browser,
                          **kwargs)
        app.initialize(argv=[])
        print
        print "***************************************************************"
        print
        print "The notebook is now live at:"
        print
        print "     http://127.0.0.1:%s/" % app.port
        print
        print "Recall you can create a new SSH tunnel dynamically by pressing"
        print "~C and then typing -L%s:localhost:%s" % (app.port, app.port)
        print "where the first number is the port on your local machine. "
        print
        print "If you are using %s on your machine already, try -L8889:localhost:%s" % (app.port, app.port)
        print
        print "Additionally, while in the notebook, we recommend you start by"
        print "replacing 'yt.mods' with 'yt.imods' like so:"
        print
        print "    from yt.imods import *"
        print
        print "This will enable some IPython-specific extensions to yt."
        print
        print "***************************************************************"
        print
        app.start()


class YTGUICmd(YTCommand):
    name = ["serve", "reason"]
    args = (
            dict(short="-o", long="--open-browser", action="store_true",
                 default = False, dest='open_browser',
                 help="Open a web browser."),
            dict(short="-p", long="--port", action="store",
                 default = 0, dest='port',
                 help="Port to listen on"),
            dict(short="-f", long="--find", action="store_true",
                 default = False, dest="find",
                 help="At startup, find all *.hierarchy files in the CWD"),
            dict(short="-d", long="--debug", action="store_true",
                 default = False, dest="debug",
                 help="Add a debugging mode for cell execution"),
            dict(short = "-r", long = "--remote", action = "store_true",
                 default = False, dest="use_pyro",
                 help = "Use with a remote Pyro4 server."),
            "opf"
            )
    description = \
        """
        Run the Web GUI Reason
        """

    def __call__(self, args):
        # We have to do a couple things.
        # First, we check that YT_DEST is set.
        if args.port == 0:
            # This means, choose one at random.  We do this by binding to a
            # socket and allowing the OS to choose the port for that socket.
            import socket
            sock = socket.socket()
            sock.bind(('', 0))
            args.port = sock.getsockname()[-1]
            del sock
        elif args.port == '-1':
            port = raw_input("Desired yt port? ")
            try:
                args.port = int(port)
            except ValueError:
                print "Please try a number next time."
                return 1
        from yt.gui.reason.utils import get_reasonjs_path
        try:
            reasonjs_path = get_reasonjs_path()
        except IOError:
            sys.exit(1)
        from yt.config import ytcfg;ytcfg["yt","__withinreason"]="True"
        import yt.utilities.bottle as bottle
        from yt.gui.reason.extdirect_repl import ExtDirectREPL
        from yt.gui.reason.bottle_mods import uuid_serve_functions, PayloadHandler
        hr = ExtDirectREPL(reasonjs_path, use_pyro=args.use_pyro)
        hr.debug = PayloadHandler.debug = args.debug
        command_line = ["pfs = []"]
        if args.find:
            # We just have to find them and store references to them.
            for fn in sorted(glob.glob("*/*.hierarchy")):
                command_line.append("pfs.append(load('%s'))" % fn[:-10])
        hr.execute("\n".join(command_line))
        bottle.debug()
        uuid_serve_functions(open_browser=args.open_browser,
                    port=int(args.port), repl=hr)

class YTStatsCmd(YTCommand):
    args = ('outputfn','bn','skip','pf','field',
            dict(long="--max", action='store_true', default=False,
                 dest='max', help="Display maximum of field requested through -f option."),
            dict(long="--min", action='store_true', default=False,
                 dest='min', help="Display minimum of field requested through -f option."))
    name = "stats"
    description = \
        """
        Print stats and max/min value of a given field (if requested),
        for one or more datasets

        (default field is Density)

        """

    def __call__(self, args):
        pf = args.pf
        pf.h.print_stats()
        if args.field in pf.h.derived_field_list:
            if args.max == True:
                v, c = pf.h.find_max(args.field)
                print "Maximum %s: %0.5e at %s" % (args.field, v, c)
            if args.min == True:
                v, c = pf.h.find_min(args.field)
                print "Minimum %s: %0.5e at %s" % (args.field, v, c)
        if args.output is not None:
            t = pf.current_time * pf['years']
            open(args.output, "a").write(
                "%s (%0.5e years): %0.5e at %s\n" % (pf, t, v, c))

class YTUpdateCmd(YTCommand):
    args = ("all", )
    name = "update"
    description = \
        """
        Update the yt installation to the most recent version

        """

    def __call__(self, opts):
        import pkg_resources
        yt_provider = pkg_resources.get_provider("yt")
        path = os.path.dirname(yt_provider.module_path)
        print
        print "yt module located at:"
        print "    %s" % (path)
        update_supp = False
        if "YT_DEST" in os.environ:
            spath = os.path.join(
                     os.environ["YT_DEST"], "src", "yt-supplemental")
            if os.path.isdir(spath):
                print "The supplemental repositories are located at:"
                print "    %s" % (spath)
                update_supp = True
        vstring = None
        if "site-packages" not in path:
            vstring = get_hg_version(path)
            print
            print "The current version of the code is:"
            print
            print "---"
            print vstring.strip()
            print "---"
            print
            print "This installation CAN be automatically updated."
            update_hg(path, skip_rebuild=opts.reinstall)
            print "Updated successfully."
            _get_yt_stack_date()
            if opts.reinstall:
                _update_yt_stack(path)
        else:
            print
            print "YT site-packages not in path, so you must"
            print "update this installation manually by committing and"
            print "merging your modifications to the code before"
            print "updating to the newest changeset."
            print

class YTUploadImageCmd(YTCommand):
    args = (dict(short="file", type=str),)
    description = \
        """
        Upload an image to imgur.com.  Must be PNG.

        """
    name = "upload_image"
    def __call__(self, args):
        filename = args.file
        if not filename.endswith(".png"):
            print "File must be a PNG file!"
            return 1
        import base64, json, pprint
        image_data = base64.b64encode(open(filename).read())
        api_key = 'f62d550859558f28c4c214136bc797c7'
        parameters = {'key':api_key, 'image':image_data, type:'base64',
                      'caption': "",
                      'title': "%s uploaded by yt" % filename}
        data = urllib.urlencode(parameters)
        req = urllib2.Request('http://api.imgur.com/2/upload.json', data)
        try:
            response = urllib2.urlopen(req).read()
        except urllib2.HTTPError as e:
            print "ERROR", e
            return {'uploaded':False}
        rv = json.loads(response)
        if 'upload' in rv and 'links' in rv['upload']:
            print
            print "Image successfully uploaded!  You can find it at:"
            print "    %s" % (rv['upload']['links']['original'])
            print
            print "If you'd like to delete it, visit this page:"
            print "    %s" % (rv['upload']['links']['delete_page'])
            print
        else:
            print
            print "Something has gone wrong!  Here is the server response:"
            print
            pprint.pprint(rv)


def run_main():
    args = parser.parse_args()
    args.func(args)

if __name__ == "__main__": run_main()



"""
Various definitions for various other modules and routines - modified
from Orion frontend.

Author: J. S. Oishi <jsoishi@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Author: Chris Malone <chris.m.malone@gmail.com>
Affiliation: SUNY Stony Brook
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2011 J.S. Oishi.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
from yt.funcs import *

parameterTypes = {"dimensionality": int,
                 "refine_by": int
                 }

# converts the Maestro inputs file name to the Enzo/yt name expected
# throughout the code. key is Maestro name, value is Enzo/yt equivalent
maestro2enzoDict = {"dm_in": "dimensionality",
                    "ref_ratio": "refine_by"
                  }

yt2maestroFieldsDict = {}
maestro2ytFieldsDict = {}

maestro_FAB_header_pattern = r"^FAB \(\((\d+), \([0-9 ]+\)\),\(\d+, \(([0-9 ]+)\)\)\)\(\((\d+,\d+,\d+)\) \((\d+,\d+,\d+)\) \((\d+,\d+,\d+)\)\) (\d+)\n"



"""
FLASH-specific data structures

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: UCSD
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import h5py
import stat
import numpy as np
import weakref

from yt.funcs import *
from yt.data_objects.grid_patch import \
    AMRGridPatch
from yt.data_objects.hierarchy import \
    AMRHierarchy
from yt.data_objects.static_output import \
    StaticOutput
from yt.utilities.definitions import \
    mpc_conversion, sec_conversion
from yt.utilities.io_handler import \
    io_registry
from yt.utilities.physical_constants import cm_per_mpc
from .fields import FLASHFieldInfo, add_flash_field, KnownFLASHFields
from yt.data_objects.field_info_container import FieldInfoContainer, NullFunc, \
     ValidateDataField, TranslationFunc

class FLASHGrid(AMRGridPatch):
    _id_offset = 1
    #__slots__ = ["_level_id", "stop_index"]
    def __init__(self, id, hierarchy, level):
        AMRGridPatch.__init__(self, id, filename = hierarchy.hierarchy_filename,
                              hierarchy = hierarchy)
        self.Parent = None
        self.Children = []
        self.Level = level

    def __repr__(self):
        return "FLASHGrid_%04i (%s)" % (self.id, self.ActiveDimensions)

class FLASHHierarchy(AMRHierarchy):

    grid = FLASHGrid
    
    def __init__(self,pf,data_style='flash_hdf5'):
        self.data_style = data_style
        self.field_indexes = {}
        self.parameter_file = weakref.proxy(pf)
        # for now, the hierarchy file is the parameter file!
        self.hierarchy_filename = self.parameter_file.parameter_filename
        self.directory = os.path.dirname(self.hierarchy_filename)
        self._handle = pf._handle
        self._particle_handle = pf._particle_handle
        
        self.float_type = np.float64
        AMRHierarchy.__init__(self,pf,data_style)

    def _initialize_data_storage(self):
        pass

    def _detect_fields(self):
        ncomp = self._handle["/unknown names"].shape[0]
        self.field_list = [s for s in self._handle["/unknown names"][:].flat]
        if ("/particle names" in self._particle_handle) :
            self.field_list += ["particle_" + s[0].strip() for s
                                in self._particle_handle["/particle names"][:]]
    
    def _setup_classes(self):
        dd = self._get_data_reader_dict()
        AMRHierarchy._setup_classes(self, dd)
        self.object_types.sort()

    def _count_grids(self):
        try:
            self.num_grids = self.parameter_file._find_parameter(
                "integer", "globalnumblocks", True)
        except KeyError:
            self.num_grids = self._handle["/simulation parameters"][0][0]
        
    def _parse_hierarchy(self):
        f = self._handle # shortcut
        pf = self.parameter_file # shortcut
        f_part = self._particle_handle # shortcut
        
        # Initialize to the domain left / domain right
        ND = self.parameter_file.dimensionality
        DLE = self.parameter_file.domain_left_edge
        DRE = self.parameter_file.domain_right_edge
        for i in range(3):
            self.grid_left_edge[:,i] = DLE[i]
            self.grid_right_edge[:,i] = DRE[i]
        # We only go up to ND for 2D datasets
        self.grid_left_edge[:,:ND] = f["/bounding box"][:,:ND,0]
        self.grid_right_edge[:,:ND] = f["/bounding box"][:,:ND,1]
        
        # Move this to the parameter file
        try:
            nxb = pf.parameters['nxb']
            nyb = pf.parameters['nyb']
            nzb = pf.parameters['nzb']
        except KeyError:
            nxb, nyb, nzb = [int(f["/simulation parameters"]['n%sb' % ax])
                              for ax in 'xyz']
        self.grid_dimensions[:] *= (nxb, nyb, nzb)
        try:
            self.grid_particle_count[:] = f_part["/localnp"][:][:,None]
        except KeyError:
            self.grid_particle_count[:] = 0.0
        self._particle_indices = np.zeros(self.num_grids + 1, dtype='int64')
        if self.num_grids > 1 :
            np.add.accumulate(self.grid_particle_count.squeeze(),
                              out=self._particle_indices[1:])
        else :
            self._particle_indices[1] = self.grid_particle_count.squeeze()
        # This will become redundant, as _prepare_grid will reset it to its
        # current value.  Note that FLASH uses 1-based indexing for refinement
        # levels, but we do not, so we reduce the level by 1.
        self.grid_levels.flat[:] = f["/refine level"][:][:] - 1
        self.grids = np.empty(self.num_grids, dtype='object')
        for i in xrange(self.num_grids):
            self.grids[i] = self.grid(i+1, self, self.grid_levels[i,0])
        

        # This is a possibly slow and verbose fix, and should be re-examined!
        rdx = (self.parameter_file.domain_width /
                self.parameter_file.domain_dimensions)
        nlevels = self.grid_levels.max()
        dxs = np.ones((nlevels+1,3),dtype='float64')
        for i in range(nlevels+1):
            dxs[i,:ND] = rdx[:ND]/self.parameter_file.refine_by**i
       
        if ND < 3:
            dxs[:,ND:] = rdx[ND:]

        for i in xrange(self.num_grids):
            dx = dxs[self.grid_levels[i],:]
            self.grid_left_edge[i][:ND] = np.rint(self.grid_left_edge[i][:ND]/dx[0][:ND])*dx[0][:ND]
            self.grid_right_edge[i][:ND] = np.rint(self.grid_right_edge[i][:ND]/dx[0][:ND])*dx[0][:ND]
                        
    def _populate_grid_objects(self):
        # We only handle 3D data, so offset is 7 (nfaces+1)
        
        offset = 7
        ii = np.argsort(self.grid_levels.flat)
        gid = self._handle["/gid"][:]
        first_ind = -(self.parameter_file.refine_by**self.parameter_file.dimensionality)
        for g in self.grids[ii].flat:
            gi = g.id - g._id_offset
            # FLASH uses 1-indexed group info
            g.Children = [self.grids[i - 1] for i in gid[gi,first_ind:] if i > -1]
            for g1 in g.Children:
                g1.Parent = g
            g._prepare_grid()
            g._setup_dx()
        if self.parameter_file.dimensionality < 3:
            DD = (self.parameter_file.domain_right_edge[2] -
                  self.parameter_file.domain_left_edge[2])
            for g in self.grids:
                g.dds[2] = DD
        if self.parameter_file.dimensionality < 2:
            DD = (self.parameter_file.domain_right_edge[1] -
                  self.parameter_file.domain_left_edge[1])
            for g in self.grids:
                g.dds[1] = DD
        self.max_level = self.grid_levels.max()

    def _setup_derived_fields(self):
        AMRHierarchy._setup_derived_fields(self)
        [self.parameter_file.conversion_factors[field] 
         for field in self.field_list]
        for field in self.field_list:
            if field not in self.derived_field_list:
                self.derived_field_list.append(field)
            if (field not in KnownFLASHFields and
                field.startswith("particle")) :
                self.parameter_file.field_info.add_field(
                        field, function=NullFunc, take_log=False,
                        validators = [ValidateDataField(field)],
                        particle_type=True)

        for field in self.derived_field_list:
            f = self.parameter_file.field_info[field]
            if f._function.func_name == "_TranslationFunc":
                # Translating an already-converted field
                self.parameter_file.conversion_factors[field] = 1.0 
                
    def _setup_data_io(self):
        self.io = io_registry[self.data_style](self.parameter_file)

class FLASHStaticOutput(StaticOutput):
    _hierarchy_class = FLASHHierarchy
    _fieldinfo_fallback = FLASHFieldInfo
    _fieldinfo_known = KnownFLASHFields
    _handle = None
    
    def __init__(self, filename, data_style='flash_hdf5',
                 storage_filename = None,
                 particle_filename = None, 
                 conversion_override = None):

        if self._handle is not None: return
        self._handle = h5py.File(filename, "r")
        if conversion_override is None: conversion_override = {}
        self._conversion_override = conversion_override
        
        self.particle_filename = particle_filename

        if self.particle_filename is None :
            self._particle_handle = self._handle
        else :
            try :
                self._particle_handle = h5py.File(self.particle_filename, "r")
            except :
                raise IOError(self.particle_filename)
                                                                
        StaticOutput.__init__(self, filename, data_style)
        self.storage_filename = storage_filename

        # These should be explicitly obtained from the file, but for now that
        # will wait until a reorganization of the source tree and better
        # generalization.
        self.refine_by = 2
        self.parameters["HydroMethod"] = 'flash' # always PPM DE
        self.parameters["Time"] = 1. # default unit is 1...
        self._set_units()
        
    def _set_units(self):
        """
        Generates the conversion to various physical _units based on the parameter file
        """
        self.units = {}
        self.time_units = {}
        if len(self.parameters) == 0:
            self._parse_parameter_file()
        self.conversion_factors = defaultdict(lambda: 1.0)
        if "EOSType" not in self.parameters:
            self.parameters["EOSType"] = -1
        if "pc_unitsbase" in self.parameters:
            if self.parameters["pc_unitsbase"] == "CGS":
                self._setup_cgs_units()
        else:
            self._setup_nounits_units()
        if self.cosmological_simulation == 1:
            self._setup_comoving_units()
        self.time_units['1'] = 1
        self.units['1'] = 1.0
        self.units['unitary'] = 1.0 / \
            (self.domain_right_edge - self.domain_left_edge).max()
        for unit in sec_conversion.keys():
            self.time_units[unit] = 1.0 / sec_conversion[unit]

        for p, v in self._conversion_override.items():
            self.conversion_factors[p] = v

    def _setup_comoving_units(self):
        self.conversion_factors['dens'] = (1.0 + self.current_redshift)**3.0
        self.conversion_factors['pres'] = (1.0 + self.current_redshift)**1.0
        self.conversion_factors['eint'] = (1.0 + self.current_redshift)**-2.0
        self.conversion_factors['ener'] = (1.0 + self.current_redshift)**-2.0
        self.conversion_factors['temp'] = (1.0 + self.current_redshift)**-2.0
        self.conversion_factors['velx'] = (1.0 + self.current_redshift)**-1.0
        self.conversion_factors['vely'] = self.conversion_factors['velx']
        self.conversion_factors['velz'] = self.conversion_factors['velx']
        self.conversion_factors['particle_velx'] = (1.0 + self.current_redshift)**-1.0
        self.conversion_factors['particle_vely'] = \
            self.conversion_factors['particle_velx']
        self.conversion_factors['particle_velz'] = \
            self.conversion_factors['particle_velx']
        if not self.has_key("TimeUnits"):
            mylog.warning("No time units.  Setting 1.0 = 1 second.")
            self.conversion_factors["Time"] = 1.0
        for unit in mpc_conversion.keys():
            self.units[unit] = mpc_conversion[unit] / mpc_conversion["cm"]
            self.units[unit+"cm"] = self.units[unit]
            self.units[unit] /= (1.0+self.current_redshift)
            
    def _setup_cgs_units(self):
        self.conversion_factors['dens'] = 1.0
        self.conversion_factors['pres'] = 1.0
        self.conversion_factors['eint'] = 1.0
        self.conversion_factors['ener'] = 1.0
        self.conversion_factors['temp'] = 1.0
        self.conversion_factors['velx'] = 1.0
        self.conversion_factors['vely'] = 1.0
        self.conversion_factors['velz'] = 1.0
        self.conversion_factors['particle_velx'] = 1.0
        self.conversion_factors['particle_vely'] = 1.0
        self.conversion_factors['particle_velz'] = 1.0
        self.conversion_factors["Time"] = 1.0
        for unit in mpc_conversion.keys():
            self.units[unit] = mpc_conversion[unit] / mpc_conversion["cm"]

    def _setup_nounits_units(self):
        self.conversion_factors['dens'] = 1.0
        self.conversion_factors['pres'] = 1.0
        self.conversion_factors['eint'] = 1.0
        self.conversion_factors['ener'] = 1.0
        self.conversion_factors['temp'] = 1.0
        self.conversion_factors['velx'] = 1.0
        self.conversion_factors['vely'] = 1.0
        self.conversion_factors['velz'] = 1.0
        self.conversion_factors['particle_velx'] = 1.0
        self.conversion_factors['particle_vely'] = 1.0
        self.conversion_factors['particle_velz'] = 1.0
        mylog.warning("Setting 1.0 in code units to be 1.0 cm")
        if not self.has_key("TimeUnits"):
            mylog.warning("No time units.  Setting 1.0 = 1 second.")
            self.conversion_factors["Time"] = 1.0
        for unit in mpc_conversion.keys():
            self.units[unit] = mpc_conversion[unit] / mpc_conversion["cm"]

    def _find_parameter(self, ptype, pname, scalar = False):
        nn = "/%s %s" % (ptype,
                {False: "runtime parameters", True: "scalars"}[scalar])
        if nn not in self._handle: raise KeyError(nn)
        for tpname, pval in zip(self._handle[nn][:,'name'],
                                self._handle[nn][:,'value']):
            if tpname.strip() == pname:
                if ptype == "string" :
                    return pval.strip()
                else :
                    return pval
        raise KeyError(pname)

    def _parse_parameter_file(self):
        self.unique_identifier = \
            int(os.stat(self.parameter_filename)[stat.ST_CTIME])
        if "file format version" in self._handle:
            self._flash_version = int(
                self._handle["file format version"][:])
        elif "sim info" in self._handle:
            self._flash_version = int(
                self._handle["sim info"][:]["file format version"])
        else:
            raise RuntimeError("Can't figure out FLASH file version.")
        # First we load all of the parameters
        hns = ["simulation parameters"]
        # note the ordering here is important: runtime parameters should
        # ovewrite scalars with the same name.
        for ptype in ['scalars', 'runtime parameters']:
            for vtype in ['integer', 'real', 'logical', 'string']:
                hns.append("%s %s" % (vtype, ptype))
        if self._flash_version > 7:
            for hn in hns:
                if hn not in self._handle:
                    continue
                for varname, val in zip(self._handle[hn][:,'name'],
                                        self._handle[hn][:,'value']):
                    vn = varname.strip()
                    if hn.startswith("string") :
                        pval = val.strip()
                    else :
                        pval = val
                    if vn in self.parameters and self.parameters[vn] != pval:
                        mylog.warning("{0} {1} overwrites a simulation scalar of the same name".format(hn[:-1],vn)) 
                    self.parameters[vn] = pval
        if self._flash_version == 7:
            for hn in hns:
                if hn not in self._handle:
                    continue
                if hn is 'simulation parameters':
                    zipover = zip(self._handle[hn].dtype.names,self._handle[hn][0])
                else:
                    zipover = zip(self._handle[hn][:,'name'],self._handle[hn][:,'value'])
                for varname, val in zipover:
                    vn = varname.strip()
                    if hn.startswith("string") :
                        pval = val.strip()
                    else :
                        pval = val
                    if vn in self.parameters and self.parameters[vn] != pval:
                        mylog.warning("{0} {1} overwrites a simulation scalar of the same name".format(hn[:-1],vn))
                    self.parameters[vn] = pval
        
        # Determine block size
        try:
            nxb = self.parameters["nxb"]
            nyb = self.parameters["nyb"]
            nzb = self.parameters["nzb"]
        except KeyError:
            nxb, nyb, nzb = [int(self._handle["/simulation parameters"]['n%sb' % ax])
                              for ax in 'xyz'] # FLASH2 only!
        
        # Determine dimensionality
        try:
            dimensionality = self.parameters["dimensionality"]
        except KeyError:
            dimensionality = 3
            if nzb == 1: dimensionality = 2
            if nyb == 1: dimensionality = 1
            if dimensionality < 3:
                mylog.warning("Guessing dimensionality as %s", dimensionality)
        
        self.dimensionality = dimensionality

        # Determine base grid parameters
        if 'lrefine_min' in self.parameters.keys() : # PARAMESH
            nblockx = self.parameters["nblockx"]
            nblocky = self.parameters["nblocky"]
            nblockz = self.parameters["nblockz"]
        else : # Uniform Grid
            nblockx = self.parameters["iprocs"]
            nblocky = self.parameters["jprocs"]
            nblockz = self.parameters["kprocs"]

        # In case the user wasn't careful
        if self.dimensionality <= 2 : nblockz = 1
        if self.dimensionality == 1 : nblocky = 1

        # Determine domain boundaries
        self.domain_left_edge = np.array(
            [self.parameters["%smin" % ax] for ax in 'xyz']).astype("float64")
        self.domain_right_edge = np.array(
            [self.parameters["%smax" % ax] for ax in 'xyz']).astype("float64")
        if self.dimensionality < 3:
            for d in (dimensionality)+range(3-dimensionality):
                if self.domain_left_edge[d] == self.domain_right_edge[d]:
                    mylog.warning('Identical domain left edge and right edges '
                                  'along dummy dimension (%i), attempting to read anyway' % d)
                    self.domain_right_edge[d] = self.domain_left_edge[d]+1.0
        self.domain_dimensions = \
            np.array([nblockx*nxb,nblocky*nyb,nblockz*nzb])

        # Try to determine Gamma
        try:
            self.parameters["Gamma"] = self.parameters["gamma"]
        except:
            mylog.warning("Cannot find Gamma")
            pass

        # Get the simulation time
        self.current_time = self.parameters["time"]

        # Determine if this is a periodic box
        p = [self.parameters.get("%sl_boundary_type" % ax, None) == "periodic" for ax in 'xyz']
        self.periodicity = tuple(p)

        # Determine cosmological parameters.
        try: 
            self.parameters["usecosmology"]
            self.cosmological_simulation = 1
            self.current_redshift = 1.0/self.parameters['scalefactor'] - 1.0
            self.omega_lambda = self.parameters['cosmologicalconstant']
            self.omega_matter = self.parameters['omegamatter']
            self.hubble_constant = self.parameters['hubbleconstant']
            self.hubble_constant *= cm_per_mpc * 1.0e-5 * 1.0e-2 # convert to 'h'
        except:
            self.current_redshift = self.omega_lambda = self.omega_matter = \
                self.hubble_constant = self.cosmological_simulation = 0.0

    def __del__(self):
        if self._handle is not self._particle_handle:
            self._particle_handle.close()
        self._handle.close()

    @classmethod
    def _is_valid(self, *args, **kwargs):
        try:
            fileh = h5py.File(args[0],'r')
            if "bounding box" in fileh["/"].keys():
                fileh.close()
                return True
            fileh.close()
        except:
            pass
        return False





"""
Matching points on the grid to specific grids

Author: John ZuHone <jzuhone@gmail.com>
Affiliation: NASA/Goddard Space Flight Center
Homepage: http://yt-project.org/
License:
Copyright (C) 2012 John ZuHone.  All Rights Reserved.

This file is part of yt.

yt is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
cimport numpy as np
cimport cython

from libc.stdlib cimport malloc, free

cdef struct GridTreeNode :
    int num_children
    int level
    int index
    np.float64_t left_edge[3]
    np.float64_t right_edge[3]
    GridTreeNode **children
                
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef GridTreeNode Grid_initialize(np.ndarray[np.float64_t, ndim=1] le,
                                  np.ndarray[np.float64_t, ndim=1] re,
                                  int num_children, int level, int index) :

    cdef GridTreeNode node
    cdef int i

    node.index = index
    node.level = level
    for i in range(3) :
        node.left_edge[i] = le[i]
        node.right_edge[i] = re[i]
    node.num_children = num_children
    
    if num_children > 0:
        node.children = <GridTreeNode **> malloc(sizeof(GridTreeNode *) *
                                                 num_children)
        for i in range(num_children) :
            node.children[i] = NULL
    else :
        node.children = NULL

    return node

cdef class GridTree :

    cdef GridTreeNode *grids
    cdef GridTreeNode *root_grids
    cdef int num_grids
    cdef int num_root_grids
    cdef int num_leaf_grids
    
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def __cinit__(self, int num_grids, 
                  np.ndarray[np.float64_t, ndim=2] left_edge,
                  np.ndarray[np.float64_t, ndim=2] right_edge,
                  np.ndarray[np.int64_t, ndim=1] parent_ind,
                  np.ndarray[np.int64_t, ndim=1] level,
                  np.ndarray[np.int64_t, ndim=1] num_children) :

        cdef int i, j, k
        cdef np.ndarray[np.int64_t, ndim=1] child_ptr

        child_ptr = np.zeros(num_grids, dtype='int64')

        self.num_grids = num_grids
        self.num_root_grids = 0
        self.num_leaf_grids = 0
        
        self.grids = <GridTreeNode *> malloc(sizeof(GridTreeNode) *
                                             num_grids)
                
        for i in range(num_grids) :

            self.grids[i] = Grid_initialize(left_edge[i,:],
                                            right_edge[i,:],
                                            num_children[i],
                                            level[i], i)
            if level[i] == 0 :
                self.num_root_grids += 1

            if num_children[i] == 0 : self.num_leaf_grids += 1

        self.root_grids = <GridTreeNode *> malloc(sizeof(GridTreeNode) *
                                                  self.num_root_grids)
                
        k = 0
        
        for i in range(num_grids) :

            j = parent_ind[i]
            
            if j >= 0:
                
                self.grids[j].children[child_ptr[j]] = &self.grids[i]

                child_ptr[j] += 1

            else :

                self.root_grids[k] = self.grids[i] 
                
                k = k + 1
    
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def return_tree_info(self) :

        cdef int i, j
        
        levels = []
        indices = []
        nchild = []
        children = []
        
        for i in range(self.num_grids) : 

            childs = []
            
            levels.append(self.grids[i].level)
            indices.append(self.grids[i].index)
            nchild.append(self.grids[i].num_children)
            for j in range(self.grids[i].num_children) :
                childs.append(self.grids[i].children[j].index)
            children.append(childs)

        return indices, levels, nchild, children
    
cdef class MatchPointsToGrids :

    cdef int num_points
    cdef np.float64_t * xp
    cdef np.float64_t * yp
    cdef np.float64_t * zp
    cdef GridTree tree
    cdef np.int64_t * point_grids

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def __cinit__(self, GridTree tree,
                  int num_points, 
                  np.ndarray[np.float64_t, ndim=1] x,
                  np.ndarray[np.float64_t, ndim=1] y,
                  np.ndarray[np.float64_t, ndim=1] z) :

        cdef int i
        
        self.num_points = num_points

        self.xp = <np.float64_t *> malloc(sizeof(np.float64_t) *
                                          num_points)
        self.yp = <np.float64_t *> malloc(sizeof(np.float64_t) *
                                          num_points)
        self.zp = <np.float64_t *> malloc(sizeof(np.float64_t) *
                                          num_points)
        self.point_grids = <np.int64_t *> malloc(sizeof(np.int64_t) *
                                              num_points)
        
        for i in range(num_points) :
            self.xp[i] = x[i]
            self.yp[i] = y[i]
            self.zp[i] = z[i]
            self.point_grids[i] = -1
            
        self.tree = tree

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def find_points_in_tree(self) :

        cdef np.ndarray[np.int64_t, ndim=1] pt_grids
        cdef int i, j
        cdef np.uint8_t in_grid
        
        pt_grids = np.zeros(self.num_points, dtype='int64')

        for i in range(self.num_points) :

            in_grid = 0
            
            for j in range(self.tree.num_root_grids) :

                if not in_grid : 
                    in_grid = self.check_position(i, self.xp[i], self.yp[i], self.zp[i],
                                                  &self.tree.root_grids[j])

        for i in range(self.num_points) :
            pt_grids[i] = self.point_grids[i]
        
        return pt_grids

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef np.uint8_t check_position(self,
                                   np.int64_t pt_index, 
                                   np.float64_t x,
                                   np.float64_t y,
                                   np.float64_t z,
                                   GridTreeNode * grid) :

        cdef int i
        cdef np.uint8_t in_grid
	
        in_grid = self.is_in_grid(x, y, z, grid)

        if in_grid :

            if grid.num_children > 0 :

                in_grid = 0
                
                for i in range(grid.num_children) :

                    if not in_grid :

                        in_grid = self.check_position(pt_index, x, y, z, grid.children[i])

                if not in_grid :
                    self.point_grids[pt_index] = grid.index
                    in_grid = 1
                    
            else :

                self.point_grids[pt_index] = grid.index
                in_grid = 1
                
        return in_grid
    
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef np.uint8_t is_in_grid(self,
			 np.float64_t x,
			 np.float64_t y,
			 np.float64_t z,
			 GridTreeNode * grid) :

        cdef np.uint8_t xcond, ycond, zcond, cond
            
        xcond = x >= grid.left_edge[0] and x < grid.right_edge[0]
        ycond = y >= grid.left_edge[1] and y < grid.right_edge[1]
        zcond = z >= grid.left_edge[2] and z < grid.right_edge[2]
	
        cond = xcond and ycond
        cond = cond and zcond

        return cond



"""
Functions to generate unique light cone solutions.

Author: Britton Smith <brittons@origins.colorado.edu>
Affiliation: CASA/University of CO, Boulder
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2011 Britton Smith.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import copy
import numpy as np
import random as rand
import sys

from yt.funcs import *
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    parallel_root_only

from .light_cone import \
     LightCone
from .common_n_volume import \
     common_volume

def project_unique_light_cones(lightcone, seed_file, field, **kwargs):
    r"""Make light cone projections using list of random
    seeds in a file.

    Parameters
    ----------
    lightcone : LightCone
        A LightCone object.
    seed_file : str
        File containing random seeds for light cone projections.
    field : str
        The field to be projected.

    Any additional parameters to project_light_cone should be
    given here.
    """

    seedList = _read_seed_file(seed_file)

    prefix = lightcone.output_prefix
    lightcone.calculate_light_cone_solution(seed=0)
    lastSeed = None

    for seed in seedList:
        if (seed['master'] != lastSeed):
            lightcone.rerandomize_light_cone_solution(seed['master'],
                                                      recycle=False)
            lastSeed = seed['master']
        if (seed['recycle'] is not None):
            lightcone.rerandomize_light_cone_solution(seed['recycle'],
                                                      recycle=True)

        lightcone.output_prefix = "%s_%s_%s" % \
          (prefix, seed['master'], seed['recycle'])
        lightcone.project_light_cone(field, **kwargs)

def find_unique_solutions(lightcone1, solutions=10, seed=None,
                          max_overlap=0.25, failures=10,
                          recycle=True, filename='unique.dat'):
    r"""Find a set of random seeds that will give light cones
    will minimal volume overlap.

    Parameters
    ----------
    lightcone1 : LightCone
        A LightCone object.
    solutions : int
        The desired number of unique solutions.
        Default: 10
    seed : int
        The random seed for generating light cone randomizations.
        Default: None
    max_overlap : float
        The maximum fractional volume an accepted solution can have
        in common with the other solutions.
        Default: 0.25
    failures : int
        The number of failed attempts before ending.
        Default: 10
    recycle : bool
        If True, attempt to use existing realizations.  This
        will speed up light cone projectinos.
        Default: True
    filename : str
        Output file name for the list of unique seeds.
        Default : 'unique.dat'

    """

    lightcone2 = copy.deepcopy(lightcone1)
    lightcone1.calculate_light_cone_solution(seed=0)
    lightcone2.calculate_light_cone_solution(seed=0)

    unique_seeds = []
    if recycle:
        master = None
    new_recycle_seed = None
    fails = 0
    recycle_fails = 0

    max_common = 0.0

    # Need to continuall save and reset the state of the random
    # number generator since it is being reset by the light
    # cone generator.
    if seed is None:
        state = None
    else:
        rand.seed(seed)
        state = rand.getstate()

    fail_digits = str(int(np.log10(failures))+1)

    while (len(unique_seeds) < solutions):
        # Create new random seed.
        if (recycle and master is not None):
            new_seed = master
            if state is not None: rand.setstate(state)
            new_recycle_seed = rand.randint(1, 1e9)
            state = rand.getstate()
        else:
            if state is not None: rand.setstate(state)
            new_seed = rand.randint(1, 1e9)
            state = rand.getstate()
            if recycle:
                master = new_seed
                recycle_fails = 0
            new_recycle_seed = None

        sys.stderr.write(("Unique solutions: %d, consecutive failures: %"+fail_digits+"d, %"+fail_digits+"d.\r") % \
                             (len(unique_seeds), fails, recycle_fails))

        lightcone1.rerandomize_light_cone_solution(new_seed,
                                                   recycle=False)
        if new_recycle_seed is not None:
            lightcone1.rerandomize_light_cone_solution(new_recycle_seed,
                                                       recycle=True)

        # Compare with all other seeds.
        test_pass = True
        for unique_seed in unique_seeds:
            lightcone2.rerandomize_light_cone_solution(unique_seed['master'],
                                                       recycle=False)
            if unique_seed['recycle'] is not None:
                lightcone2.rerandomize_light_cone_solution(unique_seed['recycle'],
                                                           recycle=True)

            common = _compare_solutions(lightcone1.light_cone_solution,
                                        lightcone2.light_cone_solution)

            if (common > max_overlap):
                test_pass = False
                break
            else:
                max_common = max(max_common, common)

        if test_pass:
            unique_seeds.append({'master':new_seed,
                                 'recycle':new_recycle_seed})
            fails = 0
            recycle_fails = 0

        else:
            if recycle:
                recycle_fails += 1
            else:
                fails += 1

            if (recycle_fails >= failures):
                sys.stderr.write(("Unique solutions: %d, consecutive failures: %"+fail_digits+"d, %"+fail_digits+"d.\n") % \
                                 (len(unique_seeds), fails, recycle_fails))
                fails += 1
                mylog.info("Max recycled failures reached with master seed %d." % \
                           new_seed)
                master = None
            if (fails >= failures):
                sys.stderr.write(("Unique solutions: %d, consecutive failures: %"+fail_digits+"d, %"+fail_digits+"d.\n") % \
                                 (len(unique_seeds), fails, recycle_fails))
                mylog.error("Max consecutive failures reached.")
                break

    mylog.info("Created %d unique solutions." % len(unique_seeds))
    mylog.info("Maximum common volume is %.2e." % max_common)
    _write_seed_file(unique_seeds, filename)
    return unique_seeds

def _compare_solutions(solution1, solution2):
    "Calculate common volume between two light cone solutions."

    if (len(solution1) != len(solution2)):
        mylog.error("Cannot compare light cone solutions with unequal numbers of slices.")
        return -1

    my_volume = 0.0
    total_volume = 0.0

    # Check that solution volumes are the same.
    if((solution1[0]['box_depth_fraction'] *
        solution1[0]['box_width_fraction']**2) !=
       (solution2[0]['box_depth_fraction'] *
        solution2[0]['box_width_fraction']**2)):
        mylog.error("Light cone solutions do not have equal volumes, will use the smaller one.")

    for q in range(len(solution1)):
        cube1 = np.zeros(shape=(len(solution1[q]['projection_center']), 2))
        volume1 = 1.0
        for w in range(len(cube1)):
            if (w == solution1[q]['projection_axis']):
                width = solution1[q]['box_depth_fraction']
            else:
                width = solution1[q]['box_width_fraction']
            volume1 *= width
            cube1[w] = [solution1[q]['projection_center'][w] - 0.5 * width,
                        solution1[q]['projection_center'][w] + 0.5 * width]

        cube2 = np.zeros(shape=(len(solution2[q]['projection_center']), 2))
        volume2 = 1.0
        for w in range(len(cube2)):
            if (w == solution2[q]['projection_axis']):
                width = solution2[q]['box_depth_fraction']
            else:
                width = solution2[q]['box_width_fraction']
            volume2 *= width
            cube2[w] = [solution2[q]['projection_center'][w] - 0.5 * width,
                        solution2[q]['projection_center'][w] + 0.5 * width]

        total_volume += min(volume1, volume2)
        my_volume += common_volume(cube1, cube2,
                                   periodic=np.array([[0, 1],
                                                      [0, 1],
                                                      [0, 1]]))

    return (my_volume / total_volume)

def _read_seed_file(filename):
    "Read list of random seeds from a file."

    mylog.info("Reading random seed list from %s." % filename)

    seed_list = []

    lines = file(filename)
    for line in lines:
        if line[0] != '#':
            line = line.strip()
            onLine = line.split(',')
            if (len(onLine) == 1):
                seed_list.append({'master':onLine[0], 'recycle':None})
            else:
                seed_list.append({'master':onLine[0], 'recycle':onLine[1]})

    return seed_list

@parallel_root_only
def _write_seed_file(seed_list, filename):
    "Write list of random seeds to a file."

    mylog.info("Writing random seed list to %s." % filename)

    f = open(filename, 'w')
    for seed in seed_list:
        if seed['recycle'] is None:
            f.write("%s\n" % seed['master'])
        else:
            f.write("%s,%s\n" % (seed['master'], seed['recycle']))
    f.close()



"""
Enzo frontend tests using moving7

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2012 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.testing import *
from yt.utilities.answer_testing.framework import \
    requires_pf, \
    small_patch_amr, \
    big_patch_amr, \
    data_dir_load
from yt.frontends.enzo.api import EnzoStaticOutput

_fields = ("Temperature", "Density", "VelocityMagnitude", "DivV",
           "particle_density")

m7 = "DD0010/moving7_0010"
@requires_pf(m7)
def test_moving7():
    pf = data_dir_load(m7)
    yield assert_equal, str(pf), "moving7_0010"
    for test in small_patch_amr(m7, _fields):
        yield test

g30 = "IsolatedGalaxy/galaxy0030/galaxy0030"
@requires_pf(g30, big_data=True)
def test_galaxy0030():
    pf = data_dir_load(g30)
    yield assert_equal, str(pf), "galaxy0030"
    for test in big_patch_amr(g30, _fields):
        yield test



"""
API for yt.frontends.art

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: UCSD
Author: J.S. Oishi <jsoishi@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Author: Britton Smith <brittonsmith@gmail.com>
Affiliation: MSU
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from .data_structures import \
      ARTGrid, \
      ARTHierarchy, \
      ARTStaticOutput

from .fields import \
      ARTFieldInfo, \
      add_art_field

from .io import \
      IOHandlerART



"""
API for yt.analysis_modules

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: UCSD
Author: J.S. Oishi <jsoishi@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Author: Britton Smith <brittonsmith@gmail.com>
Affiliation: MSU
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from .absorption_spectrum.api import \
    AbsorptionSpectrum

from .coordinate_transformation.api import \
    spherical_regrid

from .cosmological_observation.api import \
    CosmologySplice, \
    LightCone, \
    find_unique_solutions, \
    project_unique_light_cones, \
    LightRay

from .halo_finding.api import \
    Halo, \
    HOPHalo, \
    parallelHOPHalo, \
    LoadedHalo, \
    FOFHalo, \
    HaloList, \
    HOPHaloList, \
    FOFHaloList, \
    parallelHOPHaloList, \
    LoadedHaloList, \
    GenericHaloFinder, \
    parallelHF, \
    HOPHaloFinder, \
    FOFHaloFinder, \
    HaloFinder, \
    LoadHaloes

from .halo_mass_function.api import \
    HaloMassFcn, \
    TransferFunction, \
    integrate_inf

from .halo_merger_tree.api import \
    DatabaseFunctions, \
    MergerTree, \
    MergerTreeConnect, \
    Node, \
    Link, \
    MergerTreeDotOutput, \
    MergerTreeTextOutput

from .halo_profiler.api import \
    VirialFilter, \
    HaloProfiler, \
    FakeProfile

from .hierarchy_subset.api import \
    ConstructedRootGrid, \
    AMRExtractedGridProxy, \
    ExtractedHierarchy, \
    ExtractedParameterFile

from .level_sets.api import \
    coalesce_join_tree, \
    identify_contours, \
    Clump, \
    find_clumps, \
    get_lowest_clumps, \
    write_clump_hierarchy, \
    write_clumps, \
    write_old_clump_hierarchy, \
    write_old_clumps, \
    write_old_clump_info, \
    _DistanceToMainClump, \
    recursive_all_clumps, \
    return_all_clumps, \
    return_bottom_clumps, \
    recursive_bottom_clumps, \
    clump_list_sort

from .radial_column_density.api import \
    RadialColumnDensity

from .spectral_integrator.api import \
     add_xray_emissivity_field, \
     add_xray_luminosity_field, \
     add_xray_photon_emissivity_field

from .star_analysis.api import \
    StarFormationRate, \
    SpectrumBuilder

from .two_point_functions.api import \
    TwoPointFunctions, \
    FcnSet

from .radmc3d_export.api import \
    RadMC3DWriter



"""
Analyzer objects for time series datasets

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: UCSD
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import inspect

from yt.funcs import *

analysis_task_registry = {}

class AnalysisTask(object):
    class __metaclass__(type):
        def __init__(cls, name, b, d):
            type.__init__(cls, name, b, d)
            if hasattr(cls, "skip") and cls.skip == False:
                return
            analysis_task_registry[cls.__name__] = cls

    def __init__(self, *args, **kwargs):
        # This should only get called if the subclassed object
        # does not override
        if len(args) + len(kwargs) != len(self._params):
            raise RuntimeError
        self.__dict__.update(zip(self._params, args))
        self.__dict__.update(kwargs)

    def __repr__(self):
        # Stolen from AMRData.__repr__
        s = "%s: " % (self.__class__.__name__)
        s += ", ".join(["%s=%s" % (i, getattr(self,i))
                       for i in self._params])
        return s

def analysis_task(params = None):
    if params is None: params = tuple()
    def create_new_class(func):
        cls = type(func.func_name, (AnalysisTask,),
                   dict(eval = func, _params = params))
        return cls
    return create_new_class

@analysis_task(('field',))
def MaximumValue(params, data_object):
    v = data_object.quantities["MaxLocation"](
            params.field, lazy_reader=True)[0]
    return v

@analysis_task()
def CurrentTimeYears(params, pf):
    return pf.current_time * pf["years"]

class SlicePlotDataset(AnalysisTask):
    _params = ['field', 'axis', 'center']

    def __init__(self, *args, **kwargs):
        from yt.visualization.api import SlicePlot
        self.SlicePlot = SlicePlot
        AnalysisTask.__init__(self, *args, **kwargs)

    def eval(self, pf):
        slc = self.SlicePlot(pf, self.axis, self.field, center = self.center)
        return pc.save()

class QuantityProxy(AnalysisTask):
    _params = None
    quantity_name = None

    def __repr__(self):
        # Stolen from AMRData.__repr__
        s = "%s: " % (self.__class__.__name__)
        s += ", ".join(["%s" % [arg for arg in self.args]])
        s += ", ".join(["%s=%s" % (k,v) for k, v in self.kwargs.items()])
        return s

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def eval(self, data_object):
        rv = data_object.quantities[self.quantity_name](
            *self.args, **self.kwargs)
        return rv

class ParameterValue(AnalysisTask):
    _params = ['parameter']

    def __init__(self, parameter, cast=None):
        self.parameter = parameter
        if cast is None:
            cast = lambda a: a
        self.cast = cast

    def eval(self, pf):
        return self.cast(pf.get_parameter(self.parameter))

def create_quantity_proxy(quantity_object):
    args, varargs, kwargs, defaults = inspect.getargspec(quantity_object[1])
    # Strip off 'data' which is on every quantity function
    params = args[1:] 
    if kwargs is not None: params += kwargs
    dd = dict(_params = params, quantity_name = quantity_object[0])
    cls = type(quantity_object[0], (QuantityProxy,), dd)
    return cls
